# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import retryz

import storops
from storops import exception as ex
from storops.exception import get_rest_exception
from storops.lib.common import instance_cache, supplement_filesystem
from storops.unity import enums
from storops.unity.enums import FSSupportedProtocolEnum
from storops.unity.resource import UnityResource, UnityResourceList, \
    UnityAttributeResource

__author__ = 'Cedric Zhuang'


class UnityJob(UnityResource):
    @classmethod
    def create_nfs_share(cls, cli, pool, nas_server, name, size,
                         is_thin=None,
                         tiering_policy=None, async_mode=True,
                         user_cap=False):
        pool_clz = storops.unity.resource.pool.UnityPool
        nas_server_clz = storops.unity.resource.nas_server.UnityNasServer
        size = supplement_filesystem(size, user_cap)
        pool = pool_clz.get(cli, pool)
        nas_server = nas_server_clz.get(cli, nas_server)
        proto = FSSupportedProtocolEnum.NFS

        job_req_body = {
            'description': 'Creating Filesystem and share',
            'tasks': []
        }
        task_body = {
            'action': 'createFilesystem',
            'description': 'Create File System',
            'name': 'CreateNewFilesystem',
            'object': 'storageResource',
            'parametersIn': {
                'name': name,
                'description': '',
                'fsParameters': {},
                'nfsShareCreate': []
            }
        }
        fs_parameters = {
            'pool': pool,
            'nasServer': nas_server,
            'supportedProtocols': proto,
            'isThinEnabled': is_thin,
            'size': size,
            'fastVPParameters': {
                'tieringPolicy': tiering_policy
            }
        }
        nfs_share_create = {
            'name': name,
            'path': '/',

        }
        task_body['parametersIn']['fsParameters'] = cli.make_body(
            fs_parameters)
        task_body['parametersIn']['nfsShareCreate'].append(
            cli.make_body(nfs_share_create))
        job_req_body['tasks'].append(task_body)

        resp = cli.post(cls().resource_class, **job_req_body)
        resp.raise_if_err()
        job = cls(_id=resp.resource_id, cli=cli)
        if not async_mode:
            job.wait_job_completion()
        return job

    def check_errors(self):
        if self.state == enums.JobStateEnum.COMPLETED:
            return True
        elif self.state in (enums.JobStateEnum.FAILED,
                            enums.JobStateEnum.ROLLING_BACK,
                            enums.JobStateEnum.COMPLETED_WITH_ERROR):
            raise ex.JobStateError(self)
        return False

    def wait_job_completion(self, **kwargs):
        interval = kwargs.pop('interval', 5)
        timeout = kwargs.pop('timeout', 3600)

        @retryz.retry(timeout=timeout, wait=interval, on_return=False)
        def _do_update():
            self.update()
            return self.check_errors()

        try:
            _do_update()
        except retryz.RetryTimeoutError:
            raise ex.JobTimeoutException()

    @property
    @instance_cache
    def _task_messages(self):
        return [task_message
                for task in self.tasks
                for task_message in task.messages]

    @property
    @instance_cache
    def _localized_messages(self):
        return [message
                for messages in self._task_messages
                for message in messages.messages]

    @property
    @instance_cache
    def messages(self):
        return [message.message
                for message in self._localized_messages]

    @property
    @instance_cache
    def exceptions(self):
        return list(filter(lambda e: e is not None,
                           (m.to_exception() for m in self._task_messages)))

    def has_exception(self):
        return len(self.exceptions) > 0


class UnityJobList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityJob


class UnityJobTask(UnityAttributeResource):
    pass


class UnityJobTaskList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityJobTask


class UnityMessage(UnityAttributeResource):
    def to_exception(self):
        if self.error_code != 0:
            clz = get_rest_exception(self.error_code)
            ret = clz(self)
        else:
            ret = None
        return ret

    def get_messages(self):
        return [m.message for m in self.messages]


class UnityMessageList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityMessage


class UnityLocalizedMessage(UnityAttributeResource):
    pass


class UnityLocalizedMessageList(UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityLocalizedMessage


def wait_job_completion(job, **kwargs):
    interval = kwargs.pop('interval', 5)
    timeout = kwargs.pop('timeout', 3600)
    update_func = kwargs.pop('update_func', None)

    @retryz.retry(timeout=timeout, wait=interval, on_return=False)
    def _do_update(job):
        if not update_func:
            job.update()
        else:
            job = update_func(job)

        if job.state == enums.JobStateEnum.COMPLETED:
            return job
        elif job.state in (enums.JobStateEnum.FAILED,
                           enums.JobStateEnum.ROLLING_BACK,
                           enums.JobStateEnum.COMPLETED_WITH_ERROR):
            raise ex.JobStateError(job)
        return False

    try:
        return _do_update(job)
    except retryz.RetryTimeoutError:
        raise ex.JobTimeoutException()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import pytest
from hamcrest import assert_that, equal_to, only_contains

from storops import FileInterfaceRoleEnum

__author__ = 'Cedric Zhuang'


def test_check_nas_server_create_success(unity_gf):
    assert_that(unity_gf.nas_server.existed, equal_to(True))


def test_check_cifs_server_create_success(unity_gf):
    assert_that(unity_gf.nas_server.get_cifs_server().existed, equal_to(True))


def test_check_nfs_server_create_success(unity_gf):
    nfs_servers = unity_gf.unity.get_nfs_server()
    for nfs_server in nfs_servers:
        nas_server = nfs_server.nas_server
        if nas_server and nas_server.id == unity_gf.nas_server.id:
            break
    else:
        pytest.fail('cannot find the nfs server for nas server: {}'
                    .format(unity_gf.nas_server))


def test_create_delete_file_interface(unity_gf):
    ip = '2.2.2.2'
    fi = unity_gf.nas_server.create_file_interface(
        'spb_eth2', ip, role=FileInterfaceRoleEnum.PRODUCTION)
    assert_that(fi.existed, equal_to(True))
    assert_that(fi.ip_address, equal_to(ip))
    fi.delete()


def test_create_delete_dns_server(unity_gf):
    ip = '1.1.1.1'
    fi = unity_gf.nas_server.create_file_interface(
        'spa_eth2', ip, role=FileInterfaceRoleEnum.PRODUCTION)
    dns = unity_gf.nas_server.create_dns_server('test.dev', ip)
    assert_that(dns.existed, equal_to(True))
    assert_that(dns.addresses, only_contains(ip))
    dns.delete()
    fi.delete()



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, has_item, raises, instance_of, \
    none, is_not, has_items, only_contains, close_to, greater_than, not_none

from storops.exception import VNXStorageGroupError, \
    VNXStorageGroupNameInUseError, VNXDetachAluNotFoundError, \
    VNXAluAlreadyAttachedError, VNXAluNotFoundError, VNXHluNumberInUseError, \
    VNXHluAlreadyUsedError
from storops.lib.common import cache
from storops.vnx.enums import VNXSPEnum
from storops.vnx.resource.lun import VNXLun
from storops.vnx.resource.port import VNXStorageGroupHBAList
from storops.vnx.resource.sg import VNXStorageGroupList, VNXStorageGroup
from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops_test.vnx.resource.test_lun import get_lun_list

__author__ = 'Cedric Zhuang'


@patch_cli
@cache
def get_sg_list():
    return VNXStorageGroupList(t_cli())


class VNXStorageGroupListTest(TestCase):
    sg_list = get_sg_list()

    @patch_cli
    def test_get_sg_list(self):
        assert_that(self.sg_list.name,
                    has_items('VNX9495', 'ubuntu-server11', 'ubuntu-server7',
                              'ubuntu14'))
        assert_that(len(self.sg_list), equal_to(4))

    @patch_cli
    def test_poll_property(self):
        sg = self.sg_list.get('ubuntu14')
        sg.poll = False
        lun_list = sg.lun_list
        lun = lun_list[0]
        assert_that(lun_list.poll, equal_to(False))
        assert_that(lun.poll, equal_to(False))

    @patch_cli
    def test_detach_not_existed_lun(self):
        lun = VNXLun(name='y', cli=t_cli())
        # raise no error
        self.sg_list.detach_alu(lun)

    @patch_cli
    def test_get_sg_by_name(self):
        sg = self.sg_list.get('ubuntu14')
        assert_that(sg.name, equal_to('ubuntu14'))

    @patch_cli
    def test_system_lun_list_supplied(self):
        lun_list = get_lun_list()
        sgs = VNXStorageGroup.get(t_cli(), system_lun_list=lun_list)
        assert_that(len(sgs), equal_to(4))
        assert_that(lun_list.timestamp, not_none())
        for sg in sgs:
            assert_that(sg.lun_list.timestamp, equal_to(lun_list.timestamp))

    @patch_cli
    def test_sg_list_shadow_copy(self):
        filtered_sgs = self.sg_list.shadow_copy(attached_lun=15)
        assert_that(len(filtered_sgs), equal_to(1))
        assert_that(len(self.sg_list), equal_to(4))
        assert_that(filtered_sgs.timestamp, equal_to(self.sg_list.timestamp))


def get_sg(name='server7'):
    sg = VNXStorageGroup(name=name, cli=t_cli())
    sg.shuffle_hlu = False
    return sg


class VNXStorageGroupTest(TestCase):
    def setUp(self):
        self.sg_7 = get_sg()

    def tearDown(self):
        self.sg_7 = None

    @patch_cli
    def test_properties(self):
        sg = self.sg_7
        assert_that(sg.name, equal_to('server7'))
        assert_that(
            sg.wwn,
            equal_to('F6:F1:04:29:91:97:E5:11:85:E1:AE:04:FD:64:DC:17'))
        assert_that(sg.shareable, equal_to(True))
        assert_that(len(sg.alu_hlu_map), greater_than(1))
        assert_that(sg.alu_hlu_map[10], equal_to(153))
        assert_that(len(sg.hba_sp_pairs), equal_to(15))
        assert_that(sg.uid, equal_to(sg.wwn))
        assert_that(sg.has_alu(10), equal_to(True))
        assert_that(sg.has_alu(11), equal_to(False))
        assert_that(sg.has_hlu(153), equal_to(True))
        assert_that(sg.has_hlu(11), equal_to(False))
        assert_that(sg.existed, equal_to(True))

    @patch_cli
    def test_missing_two_tab(self):
        sg = VNXStorageGroup(name='a.b.c', cli=t_cli())
        assert_that(len(sg.alu_hlu_map), equal_to(1))
        assert_that(len(sg.hba_sp_pairs), equal_to(4))
        uid = 'iqn.1994-05.com.abcdef:ghijk2-l-mn-opq-2.rstuvw.xyz:c3bf2f4cca'
        port = list(filter(lambda i: i[0] == uid, sg.hba_port_list))[0][1]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(1))
        assert_that(port.host_initiator_list, has_item(uid))

    @patch_cli
    def test_missing_left_tab(self):
        sg = VNXStorageGroup(name='a.b.c', cli=t_cli())
        uid = 'iqn.1994-05.com.abcdef:ghijk2-l-mn-opq-2.rstuvw.xyz:c3bf2f4ccb'
        port = list(filter(lambda i: i[0] == uid, sg.hba_port_list))[0][1]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_B))
        assert_that(port.port_id, equal_to(1))
        assert_that(port.host_initiator_list, has_item(uid))

    @patch_cli
    def test_missing_right_tab(self):
        sg = VNXStorageGroup(name='a.b.c', cli=t_cli())
        uid = 'iqn.1994-05.com.abcdef:ghijk2-l-mn-opq-2.rstuvw.xyz:c3bf2f4ccc'
        port = list(filter(lambda i: i[0] == uid, sg.hba_port_list))[0][1]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(3))
        assert_that(port.host_initiator_list, has_item(uid))

    @patch_cli
    def test_property_hosts(self):
        hosts = get_sg('~management').hosts
        assert_that(len(hosts), equal_to(2))
        assert_that(hosts.name,
                    has_items('APM00152312055-spB', 'APM00152312055-spA'))
        assert_that(hosts[0].storage_group.name, equal_to('~management'))

    @patch_cli
    def test_property_lun_list(self):
        sg = get_sg('microsoft')
        sg.poll = False
        lun_list = sg.lun_list
        assert_that(lun_list.name, only_contains('lun4', 'lun456'))
        assert_that(lun_list.poll, equal_to(False))

    @patch_cli
    def test_property_zero_lun_list(self):
        lun_list = get_sg('os01').lun_list
        assert_that(len(lun_list), equal_to(0))

    @patch_cli
    def test_get_sg_os01(self):
        sg = VNXStorageGroup(name='os01', cli=t_cli())
        # The hba pare is incomplete, should be 0
        assert_that(len(sg.hba_sp_pairs), equal_to(0))
        assert_that(sg.hba_sp_pairs, instance_of(VNXStorageGroupHBAList))

    @patch_cli
    def test_initiator_uid_list(self):
        sg = get_sg('microsoft')
        assert_that(len(sg.initiator_uid_list), equal_to(1))
        assert_that(sg.initiator_uid_list,
                    has_item('iqn.1991-05.com.microsoft:abc.def.dev'))

    @patch_cli
    def test_hba_ports(self):
        sg = self.sg_7
        assert_that(len(sg.hba_port_list), equal_to(15))
        assert_that(len(sg.ports), equal_to(8))
        assert_that(len(sg.initiator_uid_list), equal_to(5))

    @patch_cli
    def test_iscsi_ports_all(self):
        assert_that(len(self.sg_7.iscsi_ports), equal_to(1))

    @patch_cli
    def test_fc_ports_all(self):
        assert_that(len(self.sg_7.fc_ports), equal_to(7))

    @patch_cli
    def test_get_fc_ports_with_filter_found_one(self):
        ports = self.sg_7.get_fc_ports(sp=VNXSPEnum.SP_A, port_id=2)
        assert_that(len(ports), equal_to(1))
        port = ports[0]
        assert_that(port.sp, equal_to(VNXSPEnum.SP_A))
        assert_that(port.port_id, equal_to(2))

    @patch_cli
    def test_get_iscsi_ports_with_filter_type_not_match(self):
        ports = self.sg_7.get_iscsi_ports(sp=VNXSPEnum.SP_A, port_id=2)
        assert_that(len(ports), equal_to(0))

    @patch_cli
    def test_get_port_by_sp(self):
        sg = self.sg_7
        assert_that(len(sg.get_ports(sp=VNXSPEnum.SP_A)), equal_to(5))
        assert_that(len(sg.get_ports(sp=VNXSPEnum.SP_B)), equal_to(3))
        assert_that(len(sg.get_ports(sp=VNXSPEnum.SP_A, port_id=0)),
                    equal_to(1))
        assert_that(len(sg.get_ports(sp=VNXSPEnum.SP_A, port_id=9)),
                    equal_to(0))

    @patch_cli
    def test_get_ports_by_wwn(self):
        wwn = '20:00:00:90:FA:53:4C:D0:10:00:00:90:FA:53:4C:D0'
        ports = self.sg_7.get_ports(wwn)
        assert_that(len(ports), equal_to(6))
        for port in ports:
            assert_that(port.host_initiator_list, has_item(wwn))

    @patch_cli
    def test_get_ports_by_iqn(self):
        iqn = 'iqn.2005-03.org.open-iscsi:bdbc466e1af3'
        ubuntu16 = get_sg('ubuntu16')
        ports = ubuntu16.get_ports(iqn)
        assert_that(len(ports), equal_to(3))

    @patch_cli
    def test_get_ports_no_wwn(self):
        ports = self.sg_7.get_ports()
        assert_that(len(ports), equal_to(8))

    @patch_cli
    def test_attach_alu_success(self):
        sg = self.sg_7
        lun = VNXLun(name='x', cli=t_cli())
        assert_that(sg.has_alu(0), equal_to(False))
        hlu = sg.attach_alu(lun)
        assert_that(sg.has_alu(0), equal_to(True))
        assert_that(sg.get_hlu(0), equal_to(1))
        assert_that(hlu, equal_to(1))

    @patch_cli
    def test_attach_alu_already_attached(self):
        sg = self.sg_7

        def f():
            sg.attach_alu(123)

        assert_that(f, raises(VNXAluAlreadyAttachedError,
                              'already been added'))
        assert_that(sg.get_hlu(123), equal_to(1))
        assert_that(len(sg.get_alu_hlu_map()), equal_to(3))

    @patch_cli
    def test_attach_alu_hlu_used(self):
        sg = self.sg_7

        def f():
            sg.attach_alu(123, hlu=210)

        assert_that(f, VNXHluAlreadyUsedError, 'already used')

    @patch_cli
    def test_attach_with_hlu_alu_not_found(self):
        sg = self.sg_7

        def f():
            sg.attach_alu(123, hlu=212)

        assert_that(f, VNXAluNotFoundError, 'not a bound ALU number')

    @patch_cli
    def test_attach_alu_with_hlu_success(self):
        sg = get_sg()
        hlu_id = sg.attach_alu(2, hlu=1)
        assert_that(hlu_id, equal_to(1))
        assert_that(sg.has_alu(2), equal_to(True))
        assert_that(sg.has_hlu(1), equal_to(True))

    @patch_cli
    def test_attach_alu_not_found(self):
        sg = self.sg_7

        def f():
            sg.attach_alu(124)

        assert_that(f, raises(VNXAluNotFoundError,
                              'not a bound ALU number'))
        assert_that(sg.get_hlu(124), none())
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))

    @patch_cli
    def test_attach_alu_already_attached_found_in_cache(self):
        sg = self.sg_7
        assert_that(sg.get_hlu(10), equal_to(153))
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))
        try:
            sg.attach_alu(10)
            self.fail('should have raised exception.')
        except VNXAluAlreadyAttachedError:
            assert_that(sg.get_hlu(10), equal_to(153))
            assert_that(len(sg.get_alu_hlu_map()), equal_to(2))

    @patch_cli
    def test_attach_alu_hlu_in_use_retry(self):
        def f():
            sg = self.sg_7
            sg.attach_alu(13, retry_limit=2)

        assert_that(f, raises(VNXHluNumberInUseError,
                              'LUN Number already in use'))

    @patch_cli
    def test_detach_hlu_success(self):
        sg = get_sg()
        sg.detach_alu(10)
        assert_that(sg.has_hlu(10), equal_to(False))

    @patch_cli
    def test_detach_hlu_not_found(self):
        def f():
            sg = self.sg_7
            sg.detach_alu(1032)

        assert_that(f, raises(VNXDetachAluNotFoundError, 'No such Host LUN'))

    @patch_cli
    def test_detach_hlu_not_attached(self):
        def f():
            sg = self.sg_7
            sg.detach_alu(1033)

        assert_that(f, raises(VNXDetachAluNotFoundError,
                              'is not attached'))

    @patch_cli
    def test_cache_update_incremental(self):

        sg = get_sg()
        # Append an alu/hlu into the cache.
        # make sure it's not flushed out by a sg.update call
        sg._alu_hlu_cache.update({100: 200})
        sg.update()
        sg.attach_alu(2, retry_limit=2)
        assert_that(sg.get_alu_hlu_map(),
                    equal_to({100: 200, 1032: 210, 10: 153, 2: 1}))

    @patch_cli
    def test_connect_host(self):
        def f():
            sg = self.sg_7
            sg.connect_host('host1')

        assert_that(f, raises(VNXStorageGroupError,
                              'Host specified is not known'))

    @patch_cli
    def test_disconnect_host(self):
        def f():
            sg = self.sg_7
            sg.disconnect_host('host1')

        assert_that(f, raises(VNXStorageGroupError,
                              'not currently connected'))

    @patch_cli
    def test_create_sg_name_in_use(self):
        def f():
            VNXStorageGroup.create('existed', t_cli())

        assert_that(f, raises(VNXStorageGroupNameInUseError, 'already in use'))

    @patch_cli
    def test_empty_sg_property(self):
        sg = VNXStorageGroup.get(t_cli(), 'sg1')
        wwn = 'BB:50:E8:2F:23:01:E6:11:83:36:00:60:16:58:B3:E9'
        assert_that(sg.name, equal_to('sg1'))
        assert_that(sg.wwn, equal_to(wwn))
        assert_that(len(sg.initiator_uid_list), equal_to(0))
        assert_that(len(sg.ports), equal_to(0))
        assert_that(len(sg.fc_ports), equal_to(0))
        assert_that(len(sg.iscsi_ports), equal_to(0))
        assert_that(len(sg.hba_sp_pairs), equal_to(0))

    @patch_cli
    def test_get_hlu_to_add_no_shuffle(self):

        sg = get_sg()

        sg.shuffle_hlu = False
        assert_that(sg._get_hlu_to_add(12), equal_to(1))
        sg._delete_alu(12)
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))
        assert_that(sg._get_hlu_to_add(12), equal_to(1))
        sg._delete_alu(12)
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))

    @patch_cli
    def test_get_hlu_to_add_shuffle(self):
        sg = VNXStorageGroup.get(t_cli(), 'server7')
        first = sg._get_hlu_to_add(12)
        sg._delete_alu(12)
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))
        second = sg._get_hlu_to_add(12)
        assert_that(first, is_not(equal_to(second)))
        sg._delete_alu(12)
        assert_that(len(sg.get_alu_hlu_map()), equal_to(2))

    @patch_cli
    def test_lun_list_from_shadow_copy(self):
        lun_list = get_lun_list()
        sg = VNXStorageGroup.get(cli=t_cli(), name='sg1',
                                 system_lun_list=lun_list)
        assert_that(len(sg.lun_list), equal_to(2))
        assert_that(sg.lun_list.timestamp, equal_to(lun_list.timestamp))

    sg = VNXStorageGroup.get(t_cli(), 'sg1')

    @patch_cli
    def test_sg_read_iops(self):
        assert_that(self.sg.read_iops, equal_to(3.5))

    @patch_cli
    def test_sg_write_iops(self):
        assert_that(self.sg.write_iops, equal_to(9.0))

    @patch_cli
    def test_sg_total_iops(self):
        assert_that(self.sg.total_iops, equal_to(12.5))

    @patch_cli
    def test_sg_read_mbps(self):
        assert_that(self.sg.read_mbps, equal_to(2.3 + 4.6))

    @patch_cli
    def test_sg_write_mbps(self):
        assert_that(self.sg.write_mbps, equal_to(2.7 + 5.4))

    @patch_cli
    def test_sg_total_mbps(self):
        assert_that(self.sg.total_mbps, equal_to(5.0 + 10.0))

    @patch_cli
    def test_sg_read_size_kb(self):
        assert_that(self.sg.read_size_kb, close_to(2018, 1))

    @patch_cli
    def test_sg_write_size_kb(self):
        assert_that(self.sg.write_size_kb, close_to(921, 1))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import functools
import logging
from multiprocessing.pool import ThreadPool

import six
from retryz import retry

import storops.vnx.resource.system
from storops import exception as ex
from storops.exception import OptionMissingError
from storops.lib.common import check_int, text_var, int_var, enum_var, \
    yes_no_var, list_var
from storops.lib.metric import PerfManager
from storops.vnx.enums import VNXSPEnum, VNXTieringEnum, VNXProvisionEnum, \
    VNXMigrationRate, VNXCompressionRate, \
    VNXMirrorViewRecoveryPolicy, VNXMirrorViewSyncRate, VNXLunType, \
    VNXRaidType, VNXPoolRaidType, VNXUserScopeEnum, VNXUserRoleEnum, \
    VNXCtrlMethod, VNXMirrorGroupRecoveryPolicy
from storops.vnx.heart_beat import NodeHeartBeat

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


def _get_commands(f, self, *argv, **kwargs):
    if not isinstance(self, CliClient):
        raise ValueError('self must be an instance of CliClient.')
    no_poll = not kwargs.get('poll', True)
    kwargs.pop('poll', None)
    commands = f(self, *argv, **kwargs)
    if isinstance(commands, six.string_types):
        commands = commands.split()
    commands = list(commands)
    if no_poll:
        commands.insert(0, '-np')
    return commands


def command(f):
    """ indicate it's a command of naviseccli

    :param f: function that returns the command in list
    :return: command execution result
    """

    @functools.wraps(f)
    def func_wrapper(self, *argv, **kwargs):
        if 'ip' in kwargs:
            ip = kwargs['ip']
            del kwargs['ip']
        else:
            ip = None

        commands = _get_commands(f, self, *argv, **kwargs)
        return self.execute(commands, ip=ip)

    return func_wrapper


def duel_command(f):
    """ indicate it's a command need to be called on both SP

    :param f: function that returns the command in list
    :return: command execution result on both sps (tuple of 2)
    """

    @functools.wraps(f)
    def func_wrapper(self, *argv, **kwargs):
        commands = _get_commands(f, self, *argv, **kwargs)
        return self.execute_dual(commands)

    return func_wrapper


class CliClient(PerfManager):
    def __init__(self, ip=None, username=None, password=None, scope=None,
                 sec_file=None, timeout=None, heartbeat_interval=None,
                 naviseccli=None):
        super(CliClient, self).__init__()
        if heartbeat_interval is None:
            heartbeat_interval = 60
        if scope is None:
            scope = 0
        self._heart_beat = NodeHeartBeat(
            username=username,
            password=password,
            scope=scope,
            sec_file=sec_file,
            interval=heartbeat_interval,
            timeout=timeout,
            naviseccli=naviseccli)
        self._heart_beat.add(VNXSPEnum.SP_A, ip)
        self._system_version = None

    def persist_rsc_list_metrics(self):
        persist_rsc_list = self.get_persist_rsc_list()
        if self.prev_counter and persist_rsc_list:
            for rsc_list in persist_rsc_list:
                rsc_list.persist_metric_data()

    def get_persist_rsc_list(self):
        if self.curr_counter is not None:
            ret = [rsc_list
                   for rsc_list in self.curr_counter.get_rsc_list_collection()
                   if self.is_perf_metric_enabled(rsc_list)]
        else:
            ret = []
        return ret

    def set_binary(self, binary):
        if binary is not None:
            log.info('update naviseccli binary location to: {}'.format(binary))
            self._heart_beat.set_binary(binary)

    def set_credential(self, username=None, password=None, scope=None,
                       sec_file=None):
        self._heart_beat.set_credential(username, password, scope, sec_file)

    def __del__(self):
        del self._heart_beat

    def set_ip(self, spa, spb=None, cs=None):
        self._heart_beat.add(VNXSPEnum.SP_A, spa)
        self._heart_beat.add(VNXSPEnum.SP_B, spb)
        self._heart_beat.add(VNXSPEnum.CONTROL_STATION, cs)

    @staticmethod
    def _select_one(items, allow_empty=False):
        for item in items:
            if not item:
                continue
            if item:
                ret = item
                break
        else:
            ret = []
        if not ret and not allow_empty:
            raise OptionMissingError(
                'at least one option should be available.')
        return ret

    @property
    def heartbeat(self):
        return self._heart_beat

    @command
    def get_control(self):
        return 'getcontrol'

    @command
    def get_agent(self):
        return 'getagent'

    @command
    def get_domain(self):
        return 'domain -list'

    @command
    def get_pool(self, name=None, pool_id=None):
        cmd = 'storagepool -list -all'.split()
        cmd += self._get_id_name_opt(pool_id, name, allow_empty=True)
        return cmd

    @command
    def get_lun(self, name=None, lun_id=None, lun_type=None):
        cmd = 'lun -list -all'.split()
        cmd += self._get_lun_opt(lun_id, name, lun_type, allow_empty=True)
        return cmd

    @command
    def get_cg(self, name=None):
        cmd = 'snap -group -list'.split()
        cmd += text_var('-id', name)
        cmd.append('-detail')
        return cmd

    @command
    def get_sp_port(self):
        return 'port -list -sp -all'

    @command
    def get_sp(self):
        return 'getsp'

    @command
    def get_connection_port(self, sp=None, port_id=None, vport_id=None):
        cmd = 'connection -getport -all'.split()
        if sp is not None:
            cmd += ['-sp', VNXSPEnum.get_sp_index(sp)]

        cmd += int_var('-portid', port_id)
        if port_id is not None:
            cmd += int_var('-vportid', vport_id)

        return cmd

    @command
    def get_sg(self, name=None, engineering=False):
        cmd = ['storagegroup']
        if engineering:
            cmd.append('-messner')
        cmd += '-list -host -iscsiAttributes'.split()
        cmd += text_var('-gname', name)
        return cmd

    @command
    def get_pool_feature(self):
        return ('storagepool -feature -info -isVirtualProvisioningSupported '
                '-maxPools -maxDiskDrivesPerPool -maxDiskDrivesAllPools '
                '-maxDiskDrivesPerOp -maxPoolLUNs -minPoolLUNSize '
                '-maxPoolLUNSize -numPools -numPoolLUNs -numThinLUNs '
                '-numDiskDrivesAllPools -availableDisks')

    @classmethod
    def _get_id_name_opt(cls, _id, name, allow_empty=False):
        try:
            ret = cls._select_one(
                [int_var('-id', _id), text_var('-name', name)],
                allow_empty)
        except OptionMissingError:
            raise ValueError('id or name need to be specified.')
        return ret

    @classmethod
    def _get_pool_opt(cls, pool_id, pool_name, allow_empty=False):
        try:
            ret = cls._select_one([int_var('-poolId', pool_id),
                                   text_var('-poolName', pool_name)],
                                  allow_empty=allow_empty)
        except OptionMissingError:
            raise ValueError('pool_id or pool_name need to be specified.')
        return ret

    @classmethod
    def _get_lun_opt(cls, lun_id=None, lun_name=None,
                     lun_type=None, allow_empty=False):
        try:
            ret = cls._select_one([int_var('-l', lun_id),
                                   text_var('-name', lun_name),
                                   enum_var('-showOnly', lun_type,
                                            VNXLunType)],
                                  allow_empty=allow_empty)
        except OptionMissingError:
            raise ValueError(
                'lun_id, lun_name or lun_type need to be specified.')
        return ret

    @classmethod
    def _get_primary_lun_opt(cls, primary_lun_id=None, primary_lun_name=None):
        try:
            ret = cls._select_one([int_var('-primaryLun', primary_lun_id),
                                   text_var('-primaryLunName',
                                            primary_lun_name)])
        except OptionMissingError:
            raise ValueError(
                'primary_lun_id or primary_lun_name need to be specified.')
        return ret

    @staticmethod
    def _get_provision_opt(provision):
        ret = []
        if provision is not None:
            possible_types = VNXProvisionEnum.get_all()
            if provision in possible_types:
                ret += VNXProvisionEnum.get_opt(provision)
            else:
                raise ValueError('not supported provisioning type: {}.'
                                 '  valid candidates: {}'
                                 .format(provision, possible_types))
        return ret

    @staticmethod
    def _get_tier_opt(tier):
        ret = []
        if tier is not None:
            possible_tiers = VNXTieringEnum.get_all()
            if tier in possible_tiers:
                ret += VNXTieringEnum.get_opt(tier)
            else:
                raise ValueError('not supported tiering type: {}.  '
                                 'valid candidates: {}'
                                 .format(tier, possible_tiers))
        return ret

    @command
    def create_pool_lun(self,
                        pool_name=None,
                        pool_id=None,
                        lun_name=None,
                        lun_id=None,
                        size_gb=1,
                        provision=None,
                        tier=None,
                        ignore_thresholds=None):

        cmd = ['lun', '-create', '-capacity', size_gb, '-sq', 'gb']
        cmd += self._get_pool_opt(pool_id, pool_name)
        cmd += self._get_lun_opt(lun_id, lun_name)
        cmd += self._get_provision_opt(provision)
        cmd += self._get_tier_opt(tier)
        if ignore_thresholds:
            cmd.append('-ignoreThresholds')
        return cmd

    @command
    def modify_lun(self,
                   lun_id=None,
                   lun_name=None,
                   new_name=None,
                   new_tier=None,
                   dedup=None):
        cmd = ['lun', '-modify']
        cmd += self._get_lun_opt(lun_id, lun_name)

        cmd += text_var('-newName', new_name)
        cmd += self._get_tier_opt(new_tier)

        if dedup is not None:
            dedup_op = 'on' if dedup else 'off'
            cmd += ['-deduplication', dedup_op]

        cmd.append('-o')
        return cmd

    @command
    def enable_compression(self, lun_id=None, rate=None, pool_id=None,
                           pool_name=None, ignore_thresholds=False):
        cmd = ['compression', '-on']
        cmd += int_var('-l', lun_id)
        cmd += int_var('-destPoolId', pool_id)
        cmd += text_var('-destPoolName', pool_name)
        cmd += enum_var('-rate', rate, VNXCompressionRate)
        if ignore_thresholds:
            cmd.append('-ignoreThresholds')
        cmd.append('-o')
        return cmd

    @command
    def disable_compression(self, lun_id, ignore_thresholds=False):
        cmd = ['compression', '-off']
        cmd += int_var('-l', lun_id)
        if ignore_thresholds:
            cmd.append('-ignoreThresholds')
        cmd.append('-o')
        return cmd

    @command
    def create_mount_point(self,
                           primary_lun_id=None,
                           primary_lun_name=None,
                           mount_point_name=None,
                           mount_point_id=None):
        cmd = 'lun -create -type snap'.split()
        cmd += self._get_primary_lun_opt(primary_lun_id, primary_lun_name)
        cmd += self._get_lun_opt(lun_id=mount_point_id,
                                 lun_name=mount_point_name)
        return cmd

    @command
    def attach_snap(self, snap_name, lun_id=None, lun_name=None):
        cmd = ['lun', '-attach']
        cmd += self._get_lun_opt(lun_id, lun_name)
        cmd += text_var('-snapName', snap_name)
        return cmd

    @command
    def detach_snap(self, lun_id=None, lun_name=None):
        cmd = ['lun', '-detach']
        cmd += self._get_lun_opt(lun_id, lun_name)
        cmd.append('-o')
        return cmd

    @command
    def delete_pool_lun(self,
                        lun_id=None,
                        lun_name=None,
                        delete_snapshots=False,
                        force_detach=False):
        cmd = 'lun -destroy'.split()
        cmd += self._get_lun_opt(lun_id, lun_name)

        if delete_snapshots:
            cmd.append('-destroySnapshots')

        if force_detach:
            cmd.append('-forceDetach')

        cmd.append('-o')
        return cmd

    @command
    def expand_pool_lun(self, new_size, lun_id=None, lun_name=None,
                        ignore_thresholds=False):
        cmd = ['lun', '-expand']
        cmd += self._get_lun_opt(lun_id, lun_name)
        cmd += int_var('-capacity', new_size)
        cmd += ['-sq', 'gb']
        if ignore_thresholds:
            cmd += ['-ignoreThresholds']
        cmd.append('-o')
        return cmd

    @command
    def migrate_lun(self, src_id, dst_id, rate=VNXMigrationRate.HIGH):
        cmd = ['migrate', '-start']
        cmd += int_var('-source', src_id)
        cmd += int_var('-dest', dst_id)
        cmd += enum_var('-rate', rate, VNXMigrationRate)
        cmd.append('-o')
        return cmd

    @command
    def get_migration_session(self, src_id=None):
        cmd = ['migrate', '-list']
        if src_id is not None:
            cmd += int_var('-source', src_id)
        return cmd

    @command
    def cancel_migrate_lun(self, src_id):
        if src_id is None:
            raise ValueError('source LUN id missing for cancel migration.')
        cmd = ['migrate', '-cancel']
        cmd += int_var('-source', src_id)
        cmd.append('-o')
        return cmd

    @command
    def create_sg(self, name):
        cmd = ['storagegroup', '-create']
        cmd += text_var('-gname', name)
        return cmd

    @command
    def sg_add_hlu(self, sg_name, hlu_id, alu_id):
        cmd = ['storagegroup', '-addhlu']
        cmd += int_var('-hlu', hlu_id)
        cmd += int_var('-alu', alu_id)
        cmd += text_var('-gname', sg_name)
        cmd.append('-o')
        return cmd

    @command
    def sg_delete_hlu(self, sg_name, hlu_id):
        cmd = ['storagegroup', '-removehlu']
        cmd += int_var('-hlu', hlu_id)
        cmd += text_var('-gname', sg_name)
        cmd.append('-o')
        return cmd

    @staticmethod
    def _sg_host_op(sg_name, host_name, op):
        cmd = ['storagegroup', op]
        cmd += text_var('-host', host_name)
        cmd += text_var('-gname', sg_name)
        cmd.append('-o')
        return cmd

    @command
    def sg_connect_host(self, sg_name, host_name):
        return self._sg_host_op(sg_name, host_name, '-connecthost')

    @command
    def sg_disconnect_host(self, sg_name, host_name):
        return self._sg_host_op(sg_name, host_name, '-disconnecthost')

    @command
    def config_iscsi_ip(self, sp, port_id, ip, netmask, gateway,
                        vport_id=None, vlan_id=None):
        if vport_id is None:
            vport_id = 0

        cmd = ['connection', '-setport', '-iscsi']
        cmd += ['-sp', VNXSPEnum.get_sp_index(sp)]
        cmd += int_var('-portid', port_id)
        cmd += int_var('-vportid', vport_id)
        cmd += int_var('-vlanid', vlan_id)
        cmd += text_var('-address', ip)
        cmd += text_var('-subnetmask', netmask)
        cmd += text_var('-gateway', gateway)
        cmd.append('-o')
        return cmd

    @command
    def delete_iscsi_ip(self, sp, port_id, vport_id=None):
        if vport_id is None:
            vport_id = 0

        cmd = ['connection', '-delport']
        cmd += ['-sp', VNXSPEnum.get_sp_index(sp)]
        cmd += int_var('-portid', port_id)
        cmd += int_var('-vportid', vport_id)
        cmd.append('-o')
        return cmd

    @command
    def set_path(self, sg_name, hba_uid, sp, port_id,
                 ip, host, vport_id=None):

        cmd = ['storagegroup', '-setpath']
        cmd += text_var('-gname', sg_name)
        cmd += text_var('-hbauid', hba_uid)
        cmd += ['-sp', VNXSPEnum.get_sp_index(sp)]
        cmd += int_var('-spport', port_id)
        cmd += int_var('-spvport', vport_id)
        cmd += text_var('-ip', ip)
        cmd += text_var('-host', host)
        cmd.append('-o')
        return cmd

    @command
    def delete_hba(self, hba_uid):
        return ['port', '-removeHBA', '-hbauid', hba_uid, '-o']

    @command
    def delete_sg(self, sg_name):
        cmd = ['storagegroup', '-destroy']
        cmd += text_var('-gname', sg_name)
        cmd.append('-o')
        return cmd

    @command
    def get_snap(self, name=None, res=None):
        cmd = ['snap', '-list']
        cmd += text_var('-id', name)
        cmd += int_var('-res', res)
        cmd.append('-detail')
        return cmd

    @command
    def create_snap(self, res_id, snap_name,
                    allow_rw=True, auto_delete=False, keep_for=None):
        cmd = ['snap', '-create']
        try:
            cmd += int_var('-res', res_id)
        except ValueError:
            # string type meaning cg name
            cmd += text_var('-res', res_id)
            cmd += ['-resType', 'CG']

        cmd += text_var('-name', snap_name)
        # -keepFor and -allowAutoDelete cannot co-exist
        if keep_for:
            cmd += text_var('-keepFor', keep_for)
        else:
            cmd += yes_no_var('-allowAutoDelete', auto_delete)
        cmd += yes_no_var('-allowReadWrite', allow_rw)

        return cmd

    @command
    def copy_snap(self, src_name, tgt_name,
                  ignore_migration_check=False,
                  ignore_dedup_check=False):
        cmd = ['snap', '-copy']
        cmd += text_var('-id', src_name)
        cmd += text_var('-name', tgt_name)
        if ignore_migration_check:
            cmd.append('-ignoreMigrationCheck')
        if ignore_dedup_check:
            cmd.append('-ignoreDeduplicationCheck')
        return cmd

    @command
    def modify_snap(self, name, new_name=None, desc=None,
                    auto_delete=None, allow_rw=None, keep_for=None):
        opt = []
        if new_name is not None and name != new_name:
            opt += text_var('-name', new_name)
        opt += text_var('-descr', desc)
        # -keepFor and -allowAutoDelete cannot co-exist
        if keep_for:
            opt += text_var('-keepFor', keep_for)
        else:
            opt += yes_no_var('-allowAutoDelete', auto_delete)
        opt += yes_no_var('-allowReadWrite', allow_rw)
        if len(opt) > 0:
            cmd = ['snap', '-modify', '-id', name] + opt
        else:
            cmd = []
        return cmd

    @command
    def delete_snap(self, snap_name):
        cmd = ['snap', '-destroy']
        cmd += text_var('-id', snap_name)
        cmd.append('-o')
        return cmd

    @command
    def restore_snap(self, snap_name, res_id, backup_snap=None):
        cmd = ['snap', '-restore']
        cmd += text_var('-id', snap_name)
        cmd += int_var('-res', res_id)
        if backup_snap:
            cmd += text_var('-bakName', backup_snap)
        return cmd

    @staticmethod
    def _get_cg_member_repr(members):
        def member_converter(member):
            return six.text_type(check_int(member))

        return ','.join(map(member_converter, members))

    @command
    def create_cg(self, name, members=None, auto_delete=None):
        cmd = 'snap -group -create'.split()
        cmd += text_var('-name', name)
        cmd += yes_no_var('-allowSnapAutoDelete', auto_delete)

        if members is not None and len(members) > 0:
            cmd += ['-res', self._get_cg_member_repr(members)]

        return cmd

    @classmethod
    def _cg_member_op(cls, name, op, members):
        if len(members) == 0:
            cmd = []
            log.warn('no member to add to cg: {}'.format(name))
        else:
            cmd = 'snap -group {} -id'.format(op).split()
            cmd.append(name)
            cmd += ['-res', cls._get_cg_member_repr(members)]
        return cmd

    @command
    def add_cg_member(self, name, *members):
        return self._cg_member_op(name, '-addmember', members)

    @command
    def delete_cg_member(self, name, *members):
        return self._cg_member_op(name, '-rmmember', members)

    @command
    def replace_cg_member(self, name, *members):
        return self._cg_member_op(name, '-replmember', members)

    @command
    def delete_cg(self, name):
        cmd = 'snap -group -destroy'.split()
        cmd += text_var('-id', name)
        return cmd

    @command
    def get_ndu(self, name=None):
        cmd = 'ndu -list'.split()
        if name is not None:
            cmd += text_var('-name', name)
        return cmd

    @command
    def create_mirror_view(self, name, lun_id, use_write_intent_log=True):
        cmd = 'mirror -sync -create'.split()
        cmd += text_var('-name', name)
        cmd += int_var('-lun', lun_id)
        if use_write_intent_log:
            cmd.append('-usewriteintentlog')
        else:
            cmd.append('-nowriteintentlog')
        cmd.append('-o')
        return cmd

    @command
    def delete_mirror_view(self, name):
        cmd = 'mirror -sync -destroy'.split()
        cmd += text_var('-name', name)
        cmd.append('-o')
        return cmd

    @command
    def add_mirror_view_image(self, name, sp_ip, lun_id,
                              recovery_policy=VNXMirrorViewRecoveryPolicy.AUTO,
                              sync_rate=VNXMirrorViewSyncRate.HIGH):
        cmd = 'mirror -sync -addimage'.split()
        cmd += text_var('-name', name)
        cmd += text_var('-arrayhost', sp_ip)
        cmd += int_var('-lun', lun_id)
        cmd += VNXMirrorViewRecoveryPolicy.get_opt(recovery_policy)
        cmd += enum_var('-syncrate', sync_rate, VNXMirrorViewSyncRate)
        return cmd

    @staticmethod
    def _mirror_view_image_op(op, name, image_id):
        cmd = ['mirror', '-sync', op]
        cmd += text_var('-name', name)
        cmd += text_var('-imageuid', image_id)
        cmd.append('-o')
        return cmd

    @command
    def delete_mirror_view_image(self, name, image_id):
        return self._mirror_view_image_op(
            '-removeimage', name, image_id)

    @command
    def mirror_view_fracture_image(self, name, image_id):
        return self._mirror_view_image_op(
            '-fractureimage', name, image_id)

    @command
    def mirror_view_sync_image(self, name, image_id):
        return self._mirror_view_image_op(
            '-syncimage', name, image_id)

    @command
    def mirror_view_promote_image(self, name, image_id):
        return self._mirror_view_image_op(
            '-promoteimage', name, image_id)

    @command
    def get_mirror_view(self, name=None):
        cmd = 'mirror -sync -list'.split()
        cmd += text_var('-name', name)
        return cmd

    @command
    def create_mirror_group(self, name,
                            policy=VNXMirrorGroupRecoveryPolicy.AUTO,
                            description=None):
        cmd = ['mirror', '-sync', '-creategroup', '-name', name]
        cmd += text_var('-description', description)
        cmd += VNXMirrorViewRecoveryPolicy.get_opt(policy)
        cmd += ['-o']
        return cmd

    @command
    def delete_mirror_group(self, name, force=True):
        cmd = ['mirror', '-sync', '-destroygroup', '-name', name]
        cmd += ['-force'] if force else []
        cmd += ['-o']
        return cmd

    @command
    def add_to_mirror_group(self, name, mirror_name):
        cmd = ['mirror', '-sync', '-addtogroup', '-name', name,
               '-mirrorname', mirror_name]
        return cmd

    @command
    def remove_from_mirror_group(self, name, mirror_name):
        cmd = ['mirror', '-sync', '-removefromgroup', '-name', name,
               '-mirrorname', mirror_name]
        cmd += ['-force', '-o']
        return cmd

    @command
    def sync_mirror_group(self, name):
        cmd = ['mirror', '-sync', '-syncgroup', '-name', name]
        return cmd

    @command
    def fracture_mirror_group(self, name):
        cmd = ['mirror', '-sync', '-fracturegroup', '-name', name, '-o']
        return cmd

    @command
    def promote_mirror_group(self, name, promote_type=None):
        cmd = ['mirror', '-sync', '-promotegroup', '-name', name]
        cmd += ['-o']
        return cmd

    @command
    def get_mirror_group(self, name=None):
        cmd = ['mirror', '-sync', '-listgroups']
        cmd += text_var('-name', name)
        return cmd

    @command
    def create_mirror_view_async(self, name, lun_id):
        cmd = 'mirror -async -create'.split()
        cmd += text_var('-name', name)
        cmd += int_var('-lun', lun_id)
        cmd.append('-o')
        return cmd

    @command
    def delete_mirror_view_async(self, name):
        cmd = 'mirror -async -destroy'.split()
        cmd += text_var('-name', name)
        cmd.append('-o')
        return cmd

    @command
    def add_mirror_view_image_async(
        self, name, sp_ip, lun_id,
        recovery_policy=VNXMirrorViewRecoveryPolicy.AUTO,
        sync_rate=VNXMirrorViewSyncRate.HIGH
    ):
        cmd = 'mirror -async -addimage'.split()
        cmd += text_var('-name', name)
        cmd += text_var('-arrayhost', sp_ip)
        cmd += int_var('-lun', lun_id)
        cmd += VNXMirrorViewRecoveryPolicy.get_opt(recovery_policy)
        cmd += enum_var('-syncrate', sync_rate, VNXMirrorViewSyncRate)
        return cmd

    @staticmethod
    def _mirror_view_image_async_op(op, name, image_id):
        cmd = ['mirror', '-async', op]
        cmd += text_var('-name', name)
        cmd += text_var('-imageuid', image_id)
        cmd.append('-o')
        return cmd

    @command
    def delete_mirror_view_async_image(self, name, image_id):
        return self._mirror_view_image_async_op(
            '-removeimage', name, image_id)

    @command
    def mirror_view_async_fracture_image(self, name, image_id):
        return self._mirror_view_image_async_op(
            '-fractureimage', name, image_id)

    @command
    def mirror_view_async_sync_image(self, name, image_id):
        return self._mirror_view_image_async_op(
            '-syncimage', name, image_id)

    @command
    def mirror_view_async_promote_image(self, name, image_id,
                                        promote_type=None):
        if promote_type:
            return self._mirror_view_image_async_op('-promoteimage -type {}'
                                                    .format(promote_type),
                                                    name, image_id)
        else:
            return self._mirror_view_image_async_op('-promoteimage', name,
                                                    image_id)

    @command
    def get_mirror_view_async(self, name=None):
        cmd = 'mirror -async -list'.split()
        cmd += text_var('-name', name)
        return cmd

    @command
    def create_mirror_group_async(self, name,
                                  policy=VNXMirrorGroupRecoveryPolicy.AUTO,
                                  description=None):
        cmd = ['mirror', '-async', '-creategroup', '-name', name]
        cmd += text_var('-description', description)
        cmd += VNXMirrorViewRecoveryPolicy.get_opt(policy)
        cmd += ['-o']
        return cmd

    @command
    def delete_mirror_group_async(self, name, force=True):
        cmd = ['mirror', '-async', '-destroygroup', '-name', name]
        cmd += ['-force'] if force else []
        cmd += ['-o']
        return cmd

    @command
    def add_to_mirror_group_async(self, name, mirror_name):
        cmd = ['mirror', '-async', '-addtogroup', '-name', name,
               '-mirrorname', mirror_name]
        return cmd

    @command
    def remove_from_mirror_group_async(self, name, mirror_name):
        cmd = ['mirror', '-async', '-removefromgroup', '-name', name,
               '-mirrorname', mirror_name]
        cmd += ['-force', '-o']
        return cmd

    @command
    def sync_mirror_group_async(self, name):
        cmd = ['mirror', '-async', '-syncgroup', '-name', name]
        return cmd

    @command
    def fracture_mirror_group_async(self, name):
        cmd = ['mirror', '-async', '-fracturegroup', '-name', name, '-o']
        return cmd

    @command
    def promote_mirror_group_async(self, name, promote_type=None):
        cmd = ['mirror', '-async', '-promotegroup', '-name', name]
        if promote_type:
            cmd += ['-type', promote_type]
        cmd += ['-o']
        return cmd

    @command
    def get_mirror_group_async(self, name=None):
        cmd = ['mirror', '-async', '-listgroups']
        cmd += text_var('-name', name)
        return cmd

    @command
    def get_disk(self, bus=None, enclosure=None, disk=None):
        cmd = ['getdisk']
        if bus is not None and enclosure is not None and disk is not None:
            disk_index = '{}_{}_{}'.format(bus, enclosure, disk)
            cmd.append(disk_index)
        elif bus is None and enclosure is None and disk is None:
            pass
        else:
            raise ValueError('you must specify bus, enclosure and disk id'
                             ' together to retrieve a specified disk.')
        return cmd

    @command
    def get_rg(self, rg_id=None):
        cmd = ['getrg']
        cmd += int_var(None, rg_id)
        return cmd

    @command
    def create_rg(self, disks=None, rg_id=None, raid_type=None):
        if rg_id is None:
            raise ValueError('RAID group id not specified.')
        if not disks:
            raise ValueError('disks not specified.')
        if raid_type is None:
            raid_type = VNXRaidType.RAID5
        cmd = ['createrg']
        cmd += int_var(None, rg_id)
        cmd += disks
        cmd += enum_var('-raidtype', raid_type, VNXRaidType)
        cmd.append('-o')
        return cmd

    @command
    def delete_rg(self, rg_id):
        cmd = ['removerg']
        cmd += int_var(None, rg_id)
        return cmd

    @command
    def create_pool(self, name, disks, raid_type=None):
        cmd = ['storagepool', '-create', '-disks']
        cmd += disks
        cmd += enum_var('-rtype', raid_type, VNXPoolRaidType)
        cmd += text_var('-name', name)
        cmd.append('-skiprules')
        return cmd

    @command
    def delete_pool(self, name=None, pool_id=None):
        cmd = ['storagepool', '-destroy']
        cmd += self._get_id_name_opt(pool_id, name)
        cmd.append('-o')
        return cmd

    @command
    def modify_storage_pool(self, name=None, pool_id=None,
                            new_name=None):
        cmd = ['storagepool', '-modify']
        cmd += self._get_id_name_opt(pool_id, name)
        cmd += text_var('-newName', new_name)
        cmd.append('-o')
        return cmd

    @command
    def sp_network_status(self, sp):
        sp = VNXSPEnum.get_sp_index(sp)
        return 'networkadmin -get -sp {} -all'.format(sp).split()

    @duel_command
    def delete_disk(self, disk_index):
        return 'cru_on_off -messner {} 0'.format(disk_index).split()

    @duel_command
    def install_disk(self, disk_index):
        return 'cru_on_off -messner {} 1'.format(disk_index).split()

    @command
    def ping_node(self, address, sp, port_id, vport_id=None, packet_size=None,
                  count=None, timeout=None, delay=None):
        if vport_id is None:
            vport_id = 0
        cmd = ['connection', '-pingnode']
        sp = VNXSPEnum.get_sp_index(sp)
        cmd += text_var('-sp', sp)
        cmd += int_var('-portid', port_id)
        cmd += int_var('-vportid', vport_id)
        cmd += text_var('-address', address)
        cmd += int_var('-packetSize', packet_size)
        cmd += int_var('-count', count)
        cmd += int_var('-timeout', timeout)
        cmd += int_var('-delay', delay)
        return cmd

    @command
    def list_user(self, name=None):
        cmd = ['security', '-list']
        cmd += text_var('-user', name)
        cmd.append('-type')
        return cmd

    @command
    def add_user(self, name, password, scope=None, role=None):
        if scope is None:
            scope = VNXUserScopeEnum.GLOBAL
        if role is None:
            role = VNXUserRoleEnum.ADMIN

        cmd = ['security', '-adduser']
        cmd += text_var('-user', name)
        cmd += text_var('-password', password)
        cmd += enum_var('-scope', scope, VNXUserScopeEnum)
        cmd += enum_var('-role', role, VNXUserRoleEnum)
        cmd.append('-o')
        return cmd

    @command
    def delete_user(self, name, scope=None):
        if scope is None:
            scope = VNXUserScopeEnum.GLOBAL

        cmd = ['security', '-rmuser']
        cmd += text_var('-user', name)
        cmd += enum_var('-scope', scope, VNXUserScopeEnum)
        cmd.append('-o')
        return cmd

    @command
    def get_array_name(self):
        return ['arrayname']

    @command
    def set_array_name(self, new_name):
        cmd = text_var('arrayname', new_name)
        cmd.append('-o')
        return cmd

    @command
    def set_stats(self, enable=None):
        cmd = ['setstats']
        if enable is not None:
            if enable:
                cmd.append('-on')
            else:
                cmd.append('-off')
        return cmd

    @command
    def get_ioclass(self, name=None):
        cmd = ['nqm', '-ioclass', '-list']
        cmd += text_var('-name', name)
        return cmd

    @command
    def create_ioclass(
            self, name, iotype='rw', lun_ids=None, smp_names=None,
            ctrlmethod=VNXCtrlMethod.NO_CTRL, minsize=None, maxsize=None):
        cmd = ['nqm', '-ioclass', '-create', '-name', name, '-iotype', iotype]
        cmd += int_var('-minsize', minsize)
        cmd += int_var('-maxsize', maxsize)
        cmd += list_var('-luns', lun_ids)
        cmd += list_var('-snapshots', smp_names)
        cmd += VNXCtrlMethod.parse_cmd(ctrlmethod)
        return cmd

    @command
    def modify_ioclass(self, name, new_name=None, iotype=None, lun_ids=None,
                       smp_names=None, ctrlmethod=None,
                       minsize=None, maxsize=None):
        cmd = ['nqm', '-ioclass', '-modify', '-name', name]
        cmd += text_var('-newname', new_name)
        cmd += text_var('-iotype', iotype)
        cmd += int_var('-minsize', minsize)
        cmd += int_var('-maxsize', maxsize)
        # Currently we do not support reset to anyio
        # if anyio is True:
        #     cmd += text_var('-anyio')
        cmd += list_var('-luns', lun_ids)
        cmd += list_var('-snapshots', smp_names)
        # Currently we do not support clear all luns
        # if (lun_ids and len(lun_ids) == 0)
        # and (smp_names and len(smp_names) == 0):
        #     cmd += ['-nolun']
        cmd += VNXCtrlMethod.parse_cmd(ctrlmethod)
        cmd.append('-o')
        return cmd

    @command
    def delete_ioclass(self, name):
        return ['nqm', '-ioclass', '-destroy', '-name', name, '-o']

    @command
    def get_policy(self, name=None):
        cmd = ['nqm', '-policy', '-list']
        cmd += text_var('-name', name)
        return cmd

    @command
    def create_policy(
            self, name, ioclasses=None, fail_action=None, time_limit=None,
            eval_window=None):
        cmd = ['nqm', '-policy', '-create', '-name', name]
        cmd += list_var('-ioclasses', ioclasses)
        cmd += text_var('-failaction', fail_action)
        cmd += int_var('-timelimit', time_limit)
        cmd += int_var('-evalwindow', eval_window)
        return cmd

    @command
    def modify_policy(
            self, name, new_name=None, new_ioclasses=None, time_limit=None,
            fail_action=None, eval_window=None):
        cmd = ['nqm', '-policy', '-modify', '-name', name]
        cmd += text_var('-newname', new_name)
        # This will clear all existing ioclasses
        if new_ioclasses is not None and len(new_ioclasses) == 0:
            cmd.append('-noclass')
        else:
            cmd += list_var('-ioclasses', new_ioclasses)
        cmd += text_var('-failaction', fail_action)
        cmd += int_var('-timelimit', time_limit)
        cmd += int_var('-evalwindow', eval_window)
        return cmd

    @command
    def delete_policy(self, name):
        return ['nqm', '-policy', '-destroy', '-name', name, '-o']

    @command
    def run_policy(self, name):
        return ['nqm', '-run', name]

    @command
    def stop_policy(self):
        cmd = ['nqm', '-stop', '-o']
        return cmd

    @command
    def measure_policy(self, name):
        return ['nqm', '-measure', name]

    @property
    def ip(self):
        return self._heart_beat.get_alive_sp_ip()

    @retry(on_error=ex.VNXSPDownError)
    def execute(self, params, ip=None):
        if params is not None and len(params) > 0:
            if ip is None:
                ip = self.ip
            output = self.do(ip, params)
        else:
            log.info('no command to execute.  return empty.')
            output = ''
        return output

    @retry(on_error=ex.VNXDropConnectionError)
    def do(self, ip, params):
        cmd = self._heart_beat.get_cmd_prefix(ip) + params
        return self._heart_beat.execute_cmd(ip, cmd)

    def execute_dual(self, params):
        ip_list = self._heart_beat.get_all_alive_sps_ip()
        if not self._heart_beat.is_all_sps_alive():
            raise ex.VNXSPDownError(
                'this command requires all sps available.  '
                'currently alive ips are: {}.'.format(ip_list))

        output = []
        if params is not None and len(params) > 0:
            pool = ThreadPool(len(ip_list))
            output = pool.map(lambda ip: self.do(ip, params), ip_list)
        return tuple(output)

    def set_system_version(self, version):
        self._system_version = version

    @property
    def system_version(self):
        clz = storops.vnx.resource.system.VNXAgent
        if self._system_version is None:
            self._system_version = clz(self).revision
        return self._system_version



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, equal_to, contains_string, has_item, \
    only_contains, raises, instance_of, none, is_not, not_none, close_to

from storops.exception import VNXCompressionError, \
    VNXDedupError, VNXLunNotFoundError, \
    VNXLunExtendError, VNXLunExpandSizeError, VNXLunPreparingError, \
    VNXSnapNameInUseError, VNXCompressionAlreadyEnabledError, \
    VNXLunNameInUseError, VNXTargetNotReadyError, \
    VNXCreateSnapResourceNotFoundError, VNXLunInStorageGroupError, \
    VNXAttachSnapLunTypeError, VNXLunInConsistencyGroupError, \
    VNXDetachSnapLunTypeError, VNXDedupAlreadyEnabled, \
    EnumValueNotFoundError, VNXLunHasSnapMountPointError, \
    VNXLunUsedByFeatureError, VNXNameInUseError
from storops.lib.common import instance_cache, cache
from storops.vnx.enums import VNXProvisionEnum, VNXTieringEnum, \
    VNXCompressionRate, VNXSPEnum, VNXPoolRaidType
from storops.vnx.resource.lun import VNXLun, VNXLunList
from storops.vnx.resource.snap import VNXSnap
from storops_test.vnx.cli_mock import t_cli, patch_cli
from storops_test.vnx.resource.verifiers import verify_lun_0

__author__ = 'Cedric Zhuang'


@patch_cli
@cache
def get_lun_list():
    lun_list = VNXLunList(t_cli())
    return lun_list.update()


class VNXLunTest(TestCase):
    @staticmethod
    def get_lun():
        return VNXLun(lun_id=2, cli=t_cli())

    @patch_cli
    def test_lun_properties(self):
        wwn = '60:06:01:60:12:60:3D:00:95:63:38:87:9D:69:E5:11'
        lun = VNXLun(lun_id=3, cli=t_cli())
        assert_that(lun.state, equal_to('Ready'))
        assert_that(lun.wwn, equal_to(wwn))
        assert_that(lun.status, equal_to('OK(0x0)'))
        assert_that(lun.operation, equal_to('None'))
        assert_that(lun.total_capacity_gb, equal_to(500.0))
        assert_that(lun.current_owner, equal_to(VNXSPEnum.SP_A))
        assert_that(lun.default_owner, equal_to(VNXSPEnum.SP_A))
        assert_that(lun.attached_snapshot, none())
        assert_that(lun.lun_id, equal_to(3))
        assert_that(lun.name, equal_to('File_CS0_21132_0_d7'))
        assert_that(lun.pool_name, equal_to('Pool4File'))
        assert_that(lun.is_thin_lun, equal_to(True))
        assert_that(lun.is_compressed, equal_to(False))
        assert_that(lun.is_dedup, equal_to(False))
        assert_that(lun.initial_tier, equal_to('Optimize Pool'))
        assert_that(lun.tiering_policy, none())
        assert_that(lun.is_private, equal_to(False))
        assert_that(lun.user_capacity_gbs, equal_to(500.0))
        assert_that(lun.consumed_capacity_gbs, equal_to(512.249))
        assert_that(len(lun.snapshot_mount_points), equal_to(0))
        assert_that(lun.primary_lun, none())
        assert_that(lun.raid_type, equal_to(VNXPoolRaidType.RAID_MIXED))

    @patch_cli
    def test_lun_perf_counters(self):
        lun = VNXLun(lun_id=3, cli=t_cli())
        assert_that(lun.read_requests, equal_to(1))
        assert_that(lun.read_requests_sp_a, equal_to(2))
        assert_that(lun.read_requests_sp_b, equal_to(3))
        assert_that(lun.write_requests, equal_to(4))
        assert_that(lun.write_requests_sp_a, equal_to(5))
        assert_that(lun.write_requests_sp_b, equal_to(6))
        assert_that(lun.blocks_read, equal_to(7))
        assert_that(lun.blocks_read_sp_a, equal_to(8))
        assert_that(lun.blocks_read_sp_b, equal_to(9))
        assert_that(lun.blocks_written, equal_to(10))
        assert_that(lun.blocks_written_sp_a, equal_to(11))
        assert_that(lun.blocks_written_sp_b, equal_to(12))
        assert_that(lun.busy_ticks, equal_to(13))
        assert_that(lun.busy_ticks_sp_a, equal_to(14))
        assert_that(lun.busy_ticks_sp_b, equal_to(15))
        assert_that(lun.idle_ticks, equal_to(16))
        assert_that(lun.idle_ticks_sp_a, equal_to(17))
        assert_that(lun.idle_ticks_sp_b, equal_to(18))
        assert_that(lun.sum_of_outstanding_requests, equal_to(19))
        assert_that(lun.sum_of_outstanding_requests_sp_a, equal_to(20))
        assert_that(lun.sum_of_outstanding_requests_sp_b, equal_to(21))
        assert_that(lun.non_zero_request_count_arrivals, equal_to(22))
        assert_that(lun.non_zero_request_count_arrivals_sp_a, equal_to(23))
        assert_that(lun.non_zero_request_count_arrivals_sp_b, equal_to(24))
        assert_that(lun.implicit_trespasses, equal_to(25))
        assert_that(lun.implicit_trespasses_sp_a, equal_to(26))
        assert_that(lun.implicit_trespasses_sp_b, equal_to(27))
        assert_that(lun.explicit_trespasses, equal_to(28))
        assert_that(lun.explicit_trespasses_sp_a, equal_to(29))
        assert_that(lun.explicit_trespasses_sp_b, equal_to(30))
        assert_that(lun.extreme_performance, equal_to(1.96))
        assert_that(lun.performance, equal_to(5.68))
        assert_that(lun.capacity, equal_to(92.37))

    @patch_cli
    def test_lun_state_properties(self):
        lun = VNXLun(lun_id=7, cli=t_cli())
        assert_that(lun.deduplication_state, equal_to('Off'))
        assert_that(lun.deduplication_status, equal_to('OK(0x0)'))
        assert_that(lun.is_dedup, equal_to(False))

    @patch_cli
    def test_lun_status(self):
        lun = self.get_lun()
        assert_that(lun.status, equal_to('OK(0x0)'))
        wwn = '60:06:01:60:1A:50:35:00:6D:29:F1:FC:85:78:E5:11'
        assert_that(lun.wwn, equal_to(wwn))

    @patch_cli
    def test_lun_id_setter_str_input(self):
        lun = self.get_lun()
        assert_that(lun.lun_id, equal_to(2))

    @patch_cli
    def test_lun_provision_default(self):
        lun = VNXLun(lun_id=3, cli=t_cli())
        assert_that(lun.provision, equal_to(VNXProvisionEnum.THIN))

    @patch_cli
    def test_lun_provision_thin(self):
        lun = VNXLun(lun_id=3, cli=t_cli())
        assert_that(lun.provision, equal_to(VNXProvisionEnum.THIN))

    @patch_cli
    def test_lun_provision_compressed(self):
        lun = VNXLun(lun_id=1, cli=t_cli())
        assert_that(lun.provision, equal_to(VNXProvisionEnum.COMPRESSED))

    @patch_cli
    def test_lun_provision_dedup(self):
        lun = VNXLun(lun_id=4, cli=t_cli())
        assert_that(lun.provision, equal_to(VNXProvisionEnum.DEDUPED))

    def test_lun_provision_str_not_valid(self):
        def f():
            lun = VNXLun()
            # noinspection PyPropertyAccess
            lun.provision = 'invalid'

        assert_that(f, raises(AttributeError))

    @patch_cli
    def test_lun_tier_default(self):
        lun = VNXLun(lun_id=5, cli=t_cli())
        assert_that(lun.tier, equal_to(VNXTieringEnum.HIGH_AUTO))

    def test_lun_tier_invalid_str(self):
        def f():
            lun = VNXLun()
            lun.tier = 'invalid'

        assert_that(f, raises(EnumValueNotFoundError))

    def test_lun_tier_highest_available(self):
        lun = VNXLun()
        lun.tiering_policy = 'Auto Tier'
        lun.initial_tier = 'Highest Available'
        assert_that(lun.tier, equal_to(VNXTieringEnum.HIGH_AUTO))

    def test_lun_tier_auto(self):
        lun = VNXLun()
        lun.tiering_policy = 'Auto Tier'
        lun.initial_tier = 'Optimize Pool'
        assert_that(lun.tier, equal_to(VNXTieringEnum.AUTO))

    def test_lun_tier_high(self):
        lun = VNXLun()
        lun.tiering_policy = 'Highest Available'
        lun.initial_tier = 'Highest Available'
        assert_that(lun.tier, equal_to(VNXTieringEnum.HIGH))

    def test_lun_tier_low(self):
        lun = VNXLun()
        lun.tiering_policy = 'Lowest Available'
        lun.initial_tier = 'Lowest Available'
        assert_that(lun.tier, equal_to(VNXTieringEnum.LOW))

    def test_lun_tier_no_move_high_tier(self):
        lun = VNXLun()
        lun.tiering_policy = 'No Movement'
        lun.initial_tier = 'Highest Available'
        assert_that(lun.tier, equal_to(VNXTieringEnum.NO_MOVE))

    def test_lun_tier_no_move_optimize_pool(self):
        lun = VNXLun()
        lun.tiering_policy = 'No Movement'
        lun.initial_tier = 'Optimize Pool'
        assert_that(lun.tier, equal_to(VNXTieringEnum.NO_MOVE))

    @patch_cli
    def test_update(self):
        lun = self.get_lun()
        assert_that(lun.total_capacity_gb, equal_to(2.0))
        assert_that(lun.provision, equal_to(VNXProvisionEnum.THIN))
        assert_that(lun.tier, equal_to(VNXTieringEnum.HIGH_AUTO))

    @patch_cli
    def test_repr(self):
        lun = self.get_lun()
        assert_that(str(lun), contains_string('"VNXLun": {'))
        assert_that(str(lun), contains_string(
            '"current_owner": "VNXSPEnum.SP_A"'))

    @patch_cli
    def test_get_snap(self):
        lun = VNXLun(lun_id=3, cli=t_cli())
        snaps = lun.get_snap()
        assert_that(len(snaps), equal_to(2))
        for snap in snaps:
            assert_that(snap.source_luns, has_item(lun.get_id(lun)))

    @patch_cli
    def test_get_lun_by_id(self):
        lun = VNXLun(lun_id=0, cli=t_cli())
        lun.update()
        verify_lun_0(lun)

    @patch_cli
    def test_get_lun_by_name(self):
        lun = VNXLun(name='x', cli=t_cli())
        lun.update()
        verify_lun_0(lun)

    @patch_cli
    def test_create_success(self):
        lun = VNXLun.create(t_cli(),
                            pool_id=0,
                            lun_id=2,
                            size_gb=2)
        assert_that(lun.user_capacity_gbs, equal_to(2.0))

    @patch_cli
    def test_create_name_in_use(self):
        def f():
            VNXLun.create(t_cli(), pool_id=0, lun_id=3)

        assert_that(f, raises(VNXLunNameInUseError, 'already in use'))

    def test_get_lun_id_str(self):
        assert_that(VNXLun.get_id('123'), equal_to(123))

    def test_get_lun_obj_member(self):
        lun = VNXLun(lun_id=12)
        assert_that(VNXLun.get_id(lun), equal_to(12))

    @patch_cli
    def test_get_lun_obj_property(self):
        lun = VNXLun(name='x', cli=t_cli())
        assert_that(VNXLun.get_id(lun), equal_to(0))

    def test_get_lun_id_int(self):
        assert_that(VNXLun.get_id(23), equal_to(23))

    def test_get_lun_id_err(self):
        def f():
            VNXLun.get_id('abc')

        assert_that(f, raises(ValueError, 'invalid lun number'))

    @patch_cli
    def test_get_migration_session(self):
        lun = VNXLun(lun_id=0, cli=t_cli())
        ms = lun.get_migration_session()
        assert_that(ms.existed, equal_to(True))

    @patch_cli
    def test_primary_lun_none(self):
        lun = self.get_lun()
        assert_that(lun.primary_lun, none())

    @patch_cli
    def test_attached_snapshot_none(self):
        lun = self.get_lun()
        assert_that(lun.attached_snapshot, none())

    @patch_cli
    def test_attached_snapshot_invalid_lun_type(self):
        def f():
            lun = VNXLun(name='l1', cli=t_cli())
            lun.attach_snap('s1')

        assert_that(f, raises(VNXAttachSnapLunTypeError, 'Invalid LUN type.'))

    @patch_cli
    def test_detach_snap_invalid_lun_type(self):
        def f():
            lun = VNXLun(lun_id=0, cli=t_cli())
            lun.detach_snap()

        assert_that(f, raises(VNXDetachSnapLunTypeError,
                              'not a snapshot mount point'))

    @patch_cli
    def test_create_mount_point_success(self):
        lun = VNXLun(name='l1', cli=t_cli())
        m2 = lun.create_mount_point(name='m2')
        assert_that(lun.snapshot_mount_points, instance_of(VNXLunList))
        assert_that(str(lun), contains_string('"VNXLunList": ['))
        for smp in lun.snapshot_mount_points:
            assert_that(smp, instance_of(VNXLun))
            pl = smp.primary_lun
            assert_that(pl, instance_of(VNXLun))
            assert_that(pl._get_name(), equal_to('l1'))
        assert_that(m2.attached_snapshot, none())

    @patch_cli
    def test_mount_point_properties(self):
        lun = VNXLun(name='l1', cli=t_cli())
        m1 = lun.create_mount_point(name='m1')
        assert_that(m1.name, equal_to('m1'))
        assert_that(m1.lun_id, equal_to(4057))
        s1 = m1.attached_snapshot
        assert_that(s1, instance_of(VNXSnap))
        assert_that(s1._cli, equal_to(t_cli()))
        assert_that(s1._get_name(), equal_to('s1'))

    @patch_cli
    def test_create_mount_point_name_in_use(self):
        def f():
            lun = VNXLun(name='l1', cli=t_cli())
            lun.create_mount_point(name='m3')

        assert_that(f, raises(VNXLunNameInUseError, 'already in use'))

    @patch_cli
    def test_attach_snap(self):
        m1 = VNXLun(name='m1', cli=t_cli())
        s1 = VNXSnap(name='s1', cli=t_cli())
        m1.attach_snap(s1)
        m1.update()
        assert_that(m1.attached_snapshot._get_name(), equal_to('s1'))

    @patch_cli
    def test_property_instance_cache(self):
        m1 = VNXLun(name='m1', cli=t_cli())
        s1 = m1.attached_snapshot
        s2 = m1.attached_snapshot
        assert_that(hash(s1), equal_to(hash(s2)))
        m1.update()
        s3 = m1.attached_snapshot
        assert_that(hash(s3), is_not(equal_to(hash(s1))))
        assert_that(s1._cli, not_none())

    @patch_cli
    def test_change_name(self):
        lun = VNXLun(name='m1', cli=t_cli())
        lun.name = 'l1'
        assert_that(lun.name, equal_to('l1'))

    @patch_cli
    def test_change_name_not_found(self):
        def f():
            lun = VNXLun(lun_id=4000, cli=t_cli())
            lun.name = 'l1'

        assert_that(f, raises(VNXLunNotFoundError, 'may not exist'))

    @patch_cli
    def test_change_name_failed(self):
        lun = VNXLun(name='l1', cli=t_cli())
        try:
            lun.name = 'l3'
            self.fail('should have raised an exception.')
        except VNXNameInUseError:
            assert_that(lun._get_name(), equal_to('l1'))

    @patch_cli
    def test_change_tier(self):
        def f():
            lun = VNXLun(lun_id=4000, cli=t_cli())
            lun.tier = VNXTieringEnum.LOW

        assert_that(f, raises(VNXLunNotFoundError, 'may not exist'))

    @patch_cli
    def test_expand_too_large(self):
        def f():
            lun = VNXLun(lun_id=0, cli=t_cli())
            lun.expand(999999)

        assert_that(f, raises(VNXLunExtendError,
                              'capacity specified is not supported'))

    @patch_cli
    def test_expand_file_lun(self):
        def f():
            lun = VNXLun(lun_id=1, cli=t_cli())
            lun.expand(500)

        assert_that(f, raises(VNXLunExtendError,
                              'affect a File System Storage'))

    @patch_cli
    def test_expand_too_small(self):
        def f():
            lun = VNXLun(lun_id=1, cli=t_cli())
            lun.expand(1)

        assert_that(f, raises(VNXLunExpandSizeError,
                              'greater than current LUN size'))

    @patch_cli
    def test_expand_preparing(self):
        def f():
            lun = VNXLun(lun_id=1, cli=t_cli())
            lun.expand(12)

        assert_that(f, raises(VNXLunPreparingError,
                              "is 'Preparing"))

    def test_get_id(self):
        l1 = VNXLun(lun_id=11)
        assert_that(VNXLun.get_id(l1), equal_to(11))

    @patch_cli
    def test_get_id_with_update(self):
        m1 = VNXLun(name='m1', cli=t_cli())
        assert_that(VNXLun.get_id(m1), equal_to(4057))

    def test_get_id_list(self):
        l22 = VNXLun(lun_id=22)
        l23 = VNXLun(lun_id=23)
        assert_that(VNXLun.get_id_list(l22, l23), only_contains(22, 23))

    @patch_cli
    def test_enable_compression_failed(self):
        def method():
            l1 = VNXLun(lun_id=19, cli=t_cli())
            l1.enable_compression(VNXCompressionRate.HIGH)

        def prop():
            l1 = VNXLun(lun_id=19, cli=t_cli())
            l1.is_compressed = True

        assert_that(method, raises(VNXCompressionAlreadyEnabledError,
                                   'already turned on'))
        assert_that(prop, raises(VNXCompressionError, 'not installed'))

    @patch_cli
    def test_enable_compression_ignore_threshold(self):
        def f():
            l1 = VNXLun(lun_id=3, cli=t_cli())
            l1.enable_compression(VNXCompressionRate.LOW, True)

        assert_that(f, raises(VNXCompressionAlreadyEnabledError,
                              'already turned on'))

    @patch_cli
    def test_disable_compression(self):
        def method():
            l1 = VNXLun(lun_id=19, cli=t_cli())
            l1.disable_compression()

        def prop():
            l1 = VNXLun(lun_id=19, cli=t_cli())
            l1.is_compressed = False

        assert_that(method, raises(VNXCompressionError, 'not turned on'))
        assert_that(prop, raises(VNXCompressionError, 'not turned on'))

    @patch_cli
    def test_enable_dedup(self):
        def method_call():
            l1 = VNXLun(name='l1', cli=t_cli())
            l1.enable_dedup()

        def set_property():
            l1 = VNXLun(name='l1', cli=t_cli())
            l1.is_dedup = True

        assert_that(method_call, raises(VNXDedupError, 'it is migrating'))
        assert_that(set_property, raises(VNXDedupError, 'it is migrating'))

    @patch_cli
    def test_dedup_already_enabled(self):
        def f():
            l2 = VNXLun(name='l2', cli=t_cli())
            l2.enable_dedup()

        assert_that(f, raises(VNXDedupAlreadyEnabled, 'already enabled'))

    @patch_cli
    def test_dedup_enabling(self):
        def f():
            l2 = VNXLun(name='l3', cli=t_cli())
            l2.enable_dedup()

        assert_that(f, raises(VNXDedupAlreadyEnabled, 'or enabling'))

    @patch_cli
    def test_disable_dedup(self):
        def method_call():
            l1 = VNXLun(name='l1', cli=t_cli())
            l1.disable_dedup()

        def set_property():
            l1 = VNXLun(name='l1', cli=t_cli())
            l1.is_dedup = False

        assert_that(method_call, raises(VNXDedupError, 'disabled or'))
        assert_that(set_property, raises(VNXDedupError, 'disabled or'))

    @patch_cli
    def test_delete_lun_not_exists(self):
        def f():
            l1 = VNXLun(name='not_exists', cli=t_cli())
            l1.delete()

        assert_that(f, raises(VNXLunNotFoundError, 'not exist'))

    @patch_cli
    def test_delete_lun_in_storage_group(self):
        def f():
            l2 = VNXLun(name='in_sg', cli=t_cli())
            l2.delete()

        assert_that(f, raises(VNXLunInStorageGroupError, 'in a Storage Group'))

    @patch_cli
    def test_delete_lun_in_cg(self):
        def f():
            l2 = VNXLun(name='l2', cli=t_cli())
            l2.delete()

        assert_that(f, raises(VNXLunInConsistencyGroupError,
                              'member of a consistency group'))

    @patch_cli
    def test_delete_lun_has_smp_not_force(self):
        def f():
            l2 = VNXLun(name='has_smp', cli=t_cli())
            l2.delete()

        assert_that(f, raises(VNXLunHasSnapMountPointError,
                              'has snapshot mount points'))

    @patch_cli
    def test_delete_lun_has_smp_force(self):
        lun = VNXLun(lun_id=196, cli=t_cli())
        # no error raised
        lun.delete(force=True)

    @patch_cli
    def test_delete_lun_has_migration(self):
        def f():
            lun = VNXLun(name='lun_in_migration', cli=t_cli())
            lun.delete()

        assert_that(f, raises(VNXLunUsedByFeatureError,
                              'Cannot unbind LUN'))

    @patch_cli
    def test_create_snap(self):
        def f():
            l1 = VNXLun(lun_id=11, cli=t_cli())
            l1.create_snap('s1')

        assert_that(f, raises(VNXCreateSnapResourceNotFoundError,
                              'Cannot create the snapshot'))

    @patch_cli
    def test_create_snap_existed(self):
        def f():
            l1 = VNXLun(lun_id=3, cli=t_cli())
            l1.create_snap('s1')

        assert_that(f, raises(VNXSnapNameInUseError,
                              'already in use'))

    @patch_cli
    def test_migration_dst_lun_not_available(self):
        def f():
            l1 = VNXLun(lun_id=1, cli=t_cli())
            l2 = VNXLun(lun_id=2, cli=t_cli())
            l1.migrate(l2)

        assert_that(f, raises(VNXTargetNotReadyError,
                              'not available for migration'))

    @patch_cli
    def test_create_mirror_view(self):
        lun = VNXLun(lun_id=245, cli=t_cli())
        mv = lun.create_mirror_view('mv0')
        assert_that(mv.state, equal_to('Active'))

    @patch_cli
    def test_get_source_mirror_view(self):
        lun = self.get_lun()
        assert_that(len(lun.get_mirror_view(as_src=True)), equal_to(1))

    @patch_cli
    def test_get_target_mirror_view(self):
        lun = self.get_lun()
        assert_that(len(lun.get_mirror_view(as_tgt=True)), equal_to(2))

    @patch_cli
    def test_get_mirror_view_all(self):
        lun = self.get_lun()
        assert_that(len(lun.get_mirror_view(as_src=True, as_tgt=True)),
                    equal_to(3))
        assert_that(len(lun.get_mirror_view()), equal_to(3))

    @patch_cli
    def test_force_delete_lun_not_found(self):
        def f():
            lun = VNXLun(name='y', cli=t_cli())
            lun.delete(force=True)

        assert_that(f, raises(VNXLunNotFoundError, 'may not exist'))

    @property
    @instance_cache
    def lun_5(self):
        return VNXLun(lun_id=5, cli=t_cli())

    @patch_cli
    def test_lun_read_iops(self):
        assert_that(self.lun_5.read_iops, equal_to(2.0))

    @patch_cli
    def test_lun_read_iops_spa(self):
        assert_that(self.lun_5.read_iops_sp_a, equal_to(1.5))

    @patch_cli
    def test_lun_read_iops_spb(self):
        assert_that(self.lun_5.read_iops_sp_b, equal_to(0.5))

    @patch_cli
    def test_lun_write_iops(self):
        assert_that(self.lun_5.write_iops, equal_to(4.0))

    @patch_cli
    def test_lun_write_iops_spa(self):
        assert_that(self.lun_5.write_iops_sp_a, equal_to(3.0))

    @patch_cli
    def test_lun_write_iops_spb(self):
        assert_that(self.lun_5.write_iops_sp_b, equal_to(1.0))

    @patch_cli
    def test_lun_total_iops(self):
        assert_that(self.lun_5.total_iops, equal_to(6.0))

    @patch_cli
    def test_lun_read_mbps(self):
        assert_that(self.lun_5.read_mbps, equal_to(2.3))

    @patch_cli
    def test_lun_read_mbps_spa(self):
        assert_that(self.lun_5.read_mbps_sp_a, equal_to(1.1))

    @patch_cli
    def test_lun_read_mbps_spb(self):
        assert_that(self.lun_5.read_mbps_sp_b, equal_to(1.2))

    @patch_cli
    def test_lun_write_mbps(self):
        assert_that(self.lun_5.write_mbps, equal_to(2.7))

    @patch_cli
    def test_lun_write_mbps_spa(self):
        assert_that(self.lun_5.write_mbps_sp_a, equal_to(1.3))

    @patch_cli
    def test_lun_write_mbps_spb(self):
        assert_that(self.lun_5.write_mbps_sp_b, equal_to(1.4))

    @patch_cli
    def test_lun_total_mbps(self):
        assert_that(self.lun_5.total_mbps, equal_to(5.0))

    @patch_cli
    def test_lun_read_size_kb(self):
        assert_that(self.lun_5.read_size_kb, equal_to(1177.6))

    @patch_cli
    def test_lun_write_size_kb(self):
        assert_that(self.lun_5.write_size_kb, equal_to(691.2))

    @patch_cli
    def test_lun_utilization(self):
        assert_that(self.lun_5.utilization, close_to(33.33, 0.01))

    @patch_cli
    def test_lun_utilization_spa(self):
        assert_that(self.lun_5.utilization_sp_a, close_to(44.44, 0.01))

    @patch_cli
    def test_lun_utilization_spb(self):
        assert_that(self.lun_5.utilization_sp_b, close_to(16.66, 0.01))

    @patch_cli
    def test_lun_implicit_trespasses_ps(self):
        assert_that(self.lun_5.implicit_trespasses_ps, equal_to(3.0))

    @patch_cli
    def test_lun_implicit_trespasses_ps_sp_a(self):
        assert_that(self.lun_5.implicit_trespasses_ps_sp_a, equal_to(1.0))

    @patch_cli
    def test_lun_implicit_trespasses_ps_sp_b(self):
        assert_that(self.lun_5.implicit_trespasses_ps_sp_b, equal_to(2.0))

    @patch_cli
    def test_lun_explicit_trespasses_ps(self):
        assert_that(self.lun_5.explicit_trespasses_ps, equal_to(7.0))

    @patch_cli
    def test_lun_explicit_trespasses_ps_sp_a(self):
        assert_that(self.lun_5.explicit_trespasses_ps_sp_a, equal_to(3.0))

    @patch_cli
    def test_lun_explicit_trespasses_ps_sp_b(self):
        assert_that(self.lun_5.explicit_trespasses_ps_sp_b, equal_to(4.0))


class VNXLunMigrationCallbackTest(TestCase):
    @patch_cli
    def test_migration_on_complete_session_not_found(self):
        c = _Counter()

        def on_complete():
            c.increase()

        l0 = VNXLun(name='lun0', cli=t_cli())
        l1 = VNXLun(name='lun1', cli=t_cli())
        l0.migrate(l1, on_complete=on_complete).join()
        assert_that(c.x, equal_to(1))

    @patch_cli
    def test_migration_on_error(self):
        c = _Counter()

        def on_error():
            c.decrease()

        l0 = VNXLun(name='lun0', cli=t_cli())
        l2 = VNXLun(name='lun2', cli=t_cli())
        l2.migrate(l0, on_error=on_error).join()
        assert_that(c.x, equal_to(-1))


class _Counter(object):
    def __init__(self):
        self.x = 0

    def increase(self):
        self.x += 1

    def decrease(self):
        self.x -= 1


class VNXLunListTest(TestCase):
    lun_list = get_lun_list()

    @patch_cli
    def test_get_lun_list(self):
        assert_that(self.lun_list, instance_of(VNXLunList))
        assert_that(len(self.lun_list), equal_to(183))

    @patch_cli
    def test_get_lun_by_id_found(self):
        lun = self.lun_list.get(148)
        assert_that(lun.lun_id, equal_to(148))

    @patch_cli
    def test_get_lun_by_lun(self):
        lun = self.lun_list.get(VNXLun(lun_id=148, cli=t_cli()))
        assert_that(lun.lun_id, equal_to(148))

    @patch_cli
    def test_get_lun_by_id_not_found(self):
        lun = self.lun_list.get(12345)
        assert_that(lun, none())

    @patch_cli
    def test_lun_list_perf_properties(self):
        read_iops_values = set(self.lun_list.read_iops)
        assert_that(read_iops_values, has_item(2.0))

    @patch_cli
    def test_shadow_copy(self):
        ret = self.lun_list.shadow_copy(lun_ids=[186, 151, 79])
        assert_that(len(ret), equal_to(3))
        # verify the original list is not touched
        assert_that(len(self.lun_list), equal_to(183))
        assert_that(ret.timestamp, equal_to(self.lun_list.timestamp))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import logging
import re
from datetime import datetime

from storops.lib.common import clear_instance_cache
from storops.lib.converter import to_datetime
from storops.vnx.enums import VNXSPEnum
from storops.vnx.resource import VNXCliResourceList, VNXCliResource, \
    WithListPoll

__author__ = 'Cedric Zhuang'

log = logging.getLogger(__name__)


class VNXStorageProcessor(VNXCliResource):
    def __init__(self, cli, sp=None, ip=None):
        super(VNXStorageProcessor, self).__init__()
        self._cli = cli
        self._sp = sp
        self._ip = ip

    @property
    def enum(self):
        return self._sp

    @property
    def system_timestamp(self):
        return to_datetime('{} {}'.format(self.system_date, self.system_time))

    def _get_raw_resource(self):
        sps = self._cli.get_sp(poll=self.poll).split('SP B')
        out = ''
        if sps:
            if self._sp is VNXSPEnum.SP_A:
                out = sps[0]
            elif len(sps) > 1:
                out = sps[1]

        control_out = self._cli.get_control(ip=self._ip, poll=self.poll)
        return 'SP {}\n{}\n{}'.format(self._sp.index.upper(), control_out, out)


class VNXStorageProcessorList(VNXCliResourceList):
    def __init__(self, *sp_list):
        if len(sp_list) > 0:
            cli = sp_list[0]._cli
        else:
            raise ValueError('expect at least one sp.')
        super(VNXStorageProcessorList, self).__init__(cli=cli)
        self._list = sp_list
        self._cli = cli

    def with_poll(self):
        ret = WithListPoll(self)
        for rsc in self:
            rsc.poll = True
        return ret

    def with_no_poll(self):
        ret = WithListPoll(self)
        for rsc in self:
            rsc.poll = False
        return ret

    @classmethod
    def get_resource_class(cls):
        return VNXStorageProcessor

    @clear_instance_cache
    def update(self, data=None):
        for sp in self:
            sp.update()
        self.timestamp = datetime.now()

    def get(self, sp):
        if isinstance(sp, VNXStorageProcessor):
            sp = sp.enum

        for s in self:
            if s.enum == sp:
                ret = s
                break
        else:
            ret = None
        return ret


class VNXDomainNodeList(VNXCliResourceList):
    @classmethod
    def get_resource_class(cls):
        return VNXDomainNode

    def _get_raw_resource(self):
        return self._cli.get_domain(poll=self.poll)

    def get_node(self, node_id):
        for node in self:
            if node.name == node_id:
                ret = node
                break
        else:
            log.info('node "{}" not found in the vnx domain.'.format(node_id))
            ret = None
        return ret

    @staticmethod
    def get_cs_ip(serial, cli):
        dnl = VNXDomainNodeList(cli)
        dnl.with_no_poll()
        node = dnl.get_node(serial)
        if node is None or node.control_station is None:
            log.info('system {} does not has control station.'.format(serial))
            ret = None
        else:
            ret = node.control_station.ip_address
        return ret


class VNXDomainNode(VNXCliResource):
    @property
    def name(self):
        return self.node

    @property
    def spa(self):
        return self.members.spa

    @property
    def spb(self):
        return self.members.spb

    @property
    def control_station(self):
        return self.members.control_station


class VNXDomainMemberList(VNXCliResourceList):
    def _get_member(self, index):
        def filter_by_sp_name(member):
            sp = VNXSPEnum.parse(member.name)
            return sp == index

        result = list(filter(filter_by_sp_name, self.list))
        ret = None
        if len(result) > 0:
            ret = result[0]
        return ret

    @property
    def spa(self):
        return self._get_member(VNXSPEnum.SP_A)

    @property
    def spb(self):
        return self._get_member(VNXSPEnum.SP_B)

    @property
    def control_station(self):
        return self._get_member(VNXSPEnum.CONTROL_STATION)

    @classmethod
    def get_resource_class(cls):
        return VNXDomainMember

    def _get_raw_resource(self):
        return self._cli.get_domain()


class VNXDomainMember(VNXCliResource):
    @property
    def ip(self):
        ret = 'N/A'
        ip = re.findall(r'[0-9]+(?:\.[0-9]+){3}', self.ip_address)
        if len(ip) > 0:
            ret = ip[0]
        return ret

    @property
    def is_master(self):
        return 'Master' in self.ip_address

    def _get_raw_resource(self):
        raise NotImplementedError('cli does not support list domain member '
                                  'of a specified node.  '
                                  'please use VNXDomainNode instead.')


class VNXNetworkAdmin(VNXCliResource):
    def __init__(self, sp_index, cli):
        super(VNXNetworkAdmin, self).__init__()
        self._cli = cli
        self._sp = VNXSPEnum.parse(sp_index)

    def _get_raw_resource(self):
        return self._cli.sp_network_status(self._sp, poll=self.poll)

    @staticmethod
    def _get_sp_ip(sp, cli):
        sp = VNXNetworkAdmin(VNXSPEnum.parse(sp), cli)
        sp.with_no_poll()
        return sp.ip

    @classmethod
    def get_spa_ip(cls, cli):
        return cls._get_sp_ip(VNXSPEnum.SP_A, cli)

    @classmethod
    def get_spb_ip(cls, cli):
        return cls._get_sp_ip(VNXSPEnum.SP_B, cli)



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import errno
import functools
import inspect
import json
import logging
import math
import re
import sys
import threading
from enum import Enum as _Enum
from functools import partial
from functools import wraps
from os import path, makedirs, stat

import bitmath
import cachez
import six
from retryz import retry

import storops.exception

log = logging.getLogger(__name__)


class JsonPrinter(object):
    def _get_properties(self, dec=0):
        raise NotImplementedError('need a property-value dictionary')

    def get_dict_repr(self, dec=0):
        props = self._get_properties(dec)
        props = dict((k, v) for k, v in props.items() if v is not None)
        return {self.__class__.__name__: props}

    def json(self, indent=None):
        return json.dumps(self.get_dict_repr(), indent=indent, sort_keys=True)

    def __repr__(self):
        return self.json(indent=4)

    def __str__(self):
        return self.json()


class EnumList(JsonPrinter):
    def _get_properties(self, dec=0):
        super(EnumList, self)._get_properties(dec)

    def __init__(self):
        super(EnumList, self).__init__()
        self._list = []
        self._iter = None

    @classmethod
    def get_enum_class(cls):
        raise NotImplementedError('enum class of this list is not defined.')

    @classmethod
    def parse(cls, values):
        ret = None
        if values is not None:
            ret = cls()
            ret._list = [ret.get_enum_class().parse(v) for v in values]
        return ret

    def get_dict_repr(self, dec=0):
        items = [item.get_dict_repr(dec - 1) for item in self.list]
        return {self.__class__.__name__: items}

    @property
    def list(self):
        return self._list

    def __len__(self):
        return len(self.list)

    def __iter__(self):
        self._iter = self.list.__iter__()
        return self

    def next(self):
        return next(self._iter)

    def __next__(self):
        return self.next()

    def __getitem__(self, item):
        return self.list[item]


class Enum(JsonPrinter, _Enum):
    @classmethod
    def verify(cls, value, allow_none=True):
        if value is None and not allow_none:
            raise ValueError(
                'None is not allowed here for {}.'.format(cls.__name__))
        elif value is not None and not isinstance(value, cls):
            raise ValueError('{} is not an instance of {}.'
                             .format(value, cls.__name__))

    @classmethod
    def get_all(cls):
        return list(cls)

    @classmethod
    def get_opt(cls, value):
        option_map = cls.get_option_map()
        if option_map is None:
            raise NotImplementedError('Option map is not defined for {}.'
                                      .format(cls.__name__))

        ret = option_map.get(value, None)
        if ret is None:
            raise ValueError("{} is not a valid option for {}."
                             .format(value, cls.__name__))
        return ret

    @classmethod
    def parse(cls, value):
        if isinstance(value, six.string_types):
            ret = cls.from_str(value)
        elif isinstance(value, six.integer_types):
            ret = cls.from_int(value)
        elif isinstance(value, cls):
            ret = value
        elif value is None:
            ret = None
        else:
            raise ValueError(
                'not supported value type: {}.'.format(type(value)))
        return ret

    def _get_properties(self, dec=0):
        return {'value': self.value}

    def is_equal(self, value):
        if isinstance(self.value, six.string_types):
            ret = self.value.lower() == value.lower()
        elif isinstance(self.value, six.integer_types):
            ret = self.value == int(value)
        else:
            ret = self.value == value
        return ret

    @classmethod
    def from_int(cls, value):
        ret = None
        int_index = cls.get_int_index()
        if int_index is not None:
            try:
                ret = int_index[value]
            except IndexError:
                pass
        else:
            try:
                ret = next(i for i in cls.get_all() if i.is_equal(value))
            except StopIteration:
                pass
        if ret is None:
            cls._raise_invalid_value(value)
        return ret

    @classmethod
    def from_str(cls, value):
        ret = None
        if value is not None:
            for item in cls.get_all():
                if item.is_equal(value):
                    ret = item
                    break
            else:
                cls._raise_invalid_value(value)
        return ret

    @classmethod
    def _raise_invalid_value(cls, value):
        msg = '{} is not a valid value for {}.'.format(value, cls.__name__)
        log.warn(msg)
        raise storops.exception.EnumValueNotFoundError(msg)

    @classmethod
    def get_option_map(cls):
        raise NotImplementedError('should be implemented by child class.')

    @classmethod
    def get_int_index(cls):
        return None

    @classmethod
    def values(cls):
        return [m.value for m in cls.__members__.values()]

    @classmethod
    def enum_name(cls):
        return cls.__name__

    def __str__(self):
        return JsonPrinter.__str__(self)

    def __repr__(self):
        return JsonPrinter.__repr__(self)

    def __format__(self, format_spec):
        return self._member_type_.__format__(self._value_, format_spec)

    def get_dict_repr(self, dec=0):
        if dec < 0:
            ret = '{}.{}'.format(self.__class__.__name__, self.name)
        else:
            ret = super(Enum, self).get_dict_repr(dec)
        return ret


class Dict(dict):
    def __getattr__(self, item):
        try:
            # noinspection PyUnresolvedReferences
            ret = super(Dict, self).__getattr__(item)
        except AttributeError:
            if item in self:
                value = self.get(item)
            else:
                raise AttributeError(
                    "'{}' does not contain attribute '{}'".format(
                        __name__, item))
            ret = value
        return ret


cache = cachez.cache
instance_cache = cachez.instance_cache
clear_instance_cache = cachez.clear_instance_cache


def check_int(value, allow_none=False):
    def is_digit_str():
        return isinstance(value, six.string_types) and str(value).isdigit()

    def is_int():
        return isinstance(value, int)

    if value is None and allow_none:
        ret = None
    else:
        if is_int():
            pass
        elif is_digit_str():
            value = int(value)
        else:
            raise ValueError('"{}" must be an integer.'.format(value))
        ret = int(value)
    return ret


def check_enum(value, enum_class):
    if hasattr(enum_class, 'get_all'):
        get_all_func = getattr(enum_class, 'get_all')
        candidates = get_all_func()
        parsed_enum = enum_class.parse(value)
        if parsed_enum not in candidates:
            raise ValueError('"{}" is not a valid value.  '
                             'Valid values are: {}'.format(value, candidates))
        else:
            value = parsed_enum.value
    else:
        raise ValueError('{} is not a enum.'.format(enum_class))
    return value


def check_text(value):
    if not isinstance(value, six.string_types):
        raise ValueError('"{}" must be text.'.format(value))
    return value


def check_list(value):
    if not value:
        ret = []
    elif isinstance(value, list):
        ret = value
    else:
        raise ValueError('"{}" must be list or None.'.format(value))
    return ret


def daemon(func_ref, *args, **kwargs):
    if not callable(func_ref):
        raise ValueError('background only accept callable inputs.')
    t = threading.Thread(target=func_ref, args=args, kwargs=kwargs)
    t.setDaemon(True)
    t.start()
    return t


class WeightedAverage(object):
    def __init__(self, size=5):
        self._data = []
        # linear weight
        self.weight = list(range(size, 0, -1))

    def add(self, *value):
        # the first input is the latest one.
        self._data = list(value[::-1]) + self._data
        self._data = self._data[:self.size]

    @property
    def size(self):
        return len(self.weight)

    def value(self):
        total = 0.0
        weight = 0.0
        ret = 0.0
        for v, w in zip(self._data, self.weight):
            total += v * w
            weight += w
        if weight != 0.0:
            ret = total / weight

        return ret


def _init_lock():
    return threading.Lock()


class SynchronizedDecorator(object):
    lock_map_lock = _init_lock()
    lock_map = {}

    @classmethod
    def synchronized(cls, obj=None):
        """ synchronize on obj if obj is supplied.

        :param obj: the obj to lock on.  if none, lock to the function
        :return: return of the func.
        """

        def get_key(f, o):
            if o is None:
                key = hash(f)
            else:
                key = hash(o)
            return key

        def get_lock(f, o):
            key = get_key(f, o)
            if key not in cls.lock_map:
                with cls.lock_map_lock:
                    if key not in cls.lock_map:
                        cls.lock_map[key] = _init_lock()
            return cls.lock_map[key]

        def wrap(f):
            @functools.wraps(f)
            def new_func(*args, **kw):
                with get_lock(f, obj):
                    return f(*args, **kw)

            return new_func

        return wrap


synchronized = SynchronizedDecorator.synchronized


def const_seconds(value):
    return value


def to_wait(_):
    return const_seconds(30)


retry_per_30_s = partial(retry, wait=to_wait, limit=2)


def _var(func, name, value, *param):
    if value is not None:
        if name is not None:
            ret = [name, func(value, *param)]
        else:
            ret = [func(value, *param)]
    else:
        ret = []
    return ret


def _list_var(func, name, value, *param):
    value = func(value, *param)
    if value:
        ret = [name] + value
    else:
        ret = []
    return ret


text_var = functools.partial(_var, check_text)
int_var = functools.partial(_var, check_int)
enum_var = functools.partial(_var, check_enum)
yes_no_var = functools.partial(_var, lambda b: 'yes' if b else 'no')
list_var = functools.partial(_list_var, check_list)


class Credential(object):
    def __init__(self, user=None, password=None):
        self.user = user
        self.password = password


def get_clz_from_module(module_name, clz_name):
    ret = None
    if module_name not in sys.modules:
        __import__(module_name)
    module = sys.modules[module_name]
    if hasattr(module, clz_name):
        ret = getattr(module, clz_name)
    return ret


def is_valid(value):
    return value is not None and value != 'N/A' and len(value) > 0


def get_local_folder():
    ret = path.join(path.expanduser('~'), '.storops')
    assure_folder(ret)
    return ret


def assure_folder(folder):
    if not path.exists(folder):
        try:
            makedirs(folder)
        except OSError as e:
            # ignore existing error
            if e.errno != errno.EEXIST:
                raise


def get_file_size(full_path):
    stat_info = stat(full_path)
    return stat_info.st_size


def get_lock_file(name):
    lock_folder = path.join(get_local_folder(), 'file_lock')
    assure_folder(lock_folder)
    return path.join(lock_folder, name)


def get_data_file(name):
    data_folder = get_local_folder()
    assure_folder(data_folder)
    return path.join(data_folder, name)


def round_it(n_digits=3):
    def inner(func):
        @six.wraps(func)
        def _inner(*args, **kwargs):
            value = func(*args, **kwargs)
            return round(value, n_digits)

        return _inner

    return inner


round_3 = round_it(3)


def allow_omit_parentheses(func):
    @six.wraps(func)
    def inner(*args, **kwargs):
        if len(args) == 1 and inspect.isfunction(args[0]):
            # Decorator are used without parentheses, this
            # decorator add the the parentheses
            return func()(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    return inner


class RepeatedTimer(object):
    """ Timer that execute tasks repeatedly

    Referenced from MestreLion's answer on stackoverflow:
    http://stackoverflow.com/questions/2398661/schedule-a-repeating-event-in-python-3
    """

    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.function = function
        self.interval = interval
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self, as_daemon=True):
        if not self.is_running:
            self.is_running = True
            self._timer = threading.Timer(self.interval, self._run)
            self._timer.setDaemon(as_daemon)
            self._timer.start()

    def stop(self):
        if self._timer.is_alive():
            log.info('stop timer for {}.'.format(self.function))
        self._timer.cancel()
        self.is_running = False

    def is_daemon(self):
        return self._timer.isDaemon()

    def __del__(self):
        self.stop()


def all_not_none(*items):
    return all(map(lambda item: item is not None, items))


def all_not_nan(*items):
    return all(map(lambda item: not math.isnan(item), items))


def supplement_filesystem(old_size, user_cap=False):
    """Return new size accounting for the metadata."""
    new_size = old_size
    if user_cap:
        if old_size <= _GiB_to_Byte(1.5):
            new_size = _GiB_to_Byte(3)
        else:
            new_size += _GiB_to_Byte(1.5)
    return int(new_size)


def _GiB_to_Byte(size_gb):
    return bitmath.GiB(size_gb).to_Byte().value


def is_iscsi_uid(uid):
    """Validate the iSCSI initiator format.

    :param uid: format like iqn.yyyy-mm.naming-authority:unique
    """
    return uid.startswith('iqn')


def is_fc_uid(uid):
    """Validate the FC initiator format."""
    return re.match(r"(\w{2}:){15}\w{2}", uid, re.I) is not None


def try_import(import_str, default=None):
    try:
        __import__(import_str)
        return sys.modules[import_str]
    except ImportError:
        return default


def singleton(cls, *args, **kwargs):
    instance = {}

    @wraps(cls)
    def wrapper(*args, **kwargs):
        if cls not in instance:
            instance[cls] = cls(*args, **kwargs)
        return instance[cls]

    return wrapper



# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import logging

from storops.unity import resource
import storops.unity.resource.cifs_share
import storops.unity.resource.interface
import storops.unity.resource.nas_server

__author__ = 'Jay Xu'

LOG = logging.getLogger(__name__)


class UnityCifsServer(resource.UnityResource):
    @classmethod
    def create(cls, cli, nas_server, interfaces=None,
               netbios_name=None, name=None,
               domain=None, domain_username=None, domain_password=None,
               workgroup=None, local_password=None):
        if interfaces is None:
            interfaces = []
        nas_server_clz = storops.unity.resource.nas_server.UnityNasServer
        fi_clz = storops.unity.resource.interface.UnityFileInterface

        nas_server = nas_server_clz.get(cli, nas_server)
        interfaces = [fi_clz.get(cli, fi) for fi in interfaces]
        resp = cli.post(cls().resource_class,
                        nasServer=nas_server,
                        netbiosName=netbios_name,
                        name=name,
                        domain=domain,
                        domainUsername=domain_username,
                        domainPassword=domain_password,
                        interfaces=interfaces,
                        workgroup=workgroup,
                        localAdminPassword=local_password)
        resp.raise_if_err()
        return cls(_id=resp.resource_id, cli=cli)

    @classmethod
    def get(cls, cli, _id=None):
        nas_server_clz = storops.unity.resource.nas_server.UnityNasServer
        if isinstance(_id, nas_server_clz):
            ret = _id.get_cifs_server()
        elif not isinstance(_id, cls):
            ret = cls(_id=_id, cli=cli)
        else:
            ret = _id
        return ret

    def create_cifs_share(self, name, fs, path=None):
        clz = storops.unity.resource.cifs_share.UnityCifsShare
        return clz.create(self._cli, name=name, fs=fs,
                          path=path, cifs_server=self)

    def delete(self, skip_domain_unjoin=None, username=None,
               password=None, async_mode=False):
        resp = self._cli.delete(self.resource_class,
                                self.get_id(),
                                skipUnjoin=skip_domain_unjoin,
                                domainUsername=username,
                                domainPassword=password,
                                async_mode=async_mode)
        resp.raise_if_err()
        return resp


class UnityCifsServerList(resource.UnityResourceList):
    @classmethod
    def get_resource_class(cls):
        return UnityCifsServer



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, raises, equal_to, has_items, close_to

from storops.lib.common import instance_cache, cache
from storops_test.vnx.cli_mock import patch_cli, t_cli
from storops_test.vnx.resource.verifiers import verify_disk_4_0_e8
from storops.vnx.resource.disk import VNXDisk, VNXDiskList

__author__ = 'Cedric Zhuang'


@patch_cli
@cache
def get_disks():
    return VNXDisk.get(t_cli())


class VNXDiskTest(TestCase):
    def test_invalid_index(self):
        def f():
            VNXDisk('abcde')

        assert_that(f, raises(ValueError, 'invalid disk index'))

    def test_parse_index(self):
        bus, enc, disk = VNXDisk.parse_index('1_10_A4')
        assert_that(bus, equal_to("1"))
        assert_that(enc, equal_to("10"))
        assert_that(disk, equal_to("A4"))

    def test_parse_index_error(self):
        def f():
            VNXDisk.parse_index('abcdefg')

        assert_that(f, raises(ValueError, 'invalid disk index'))

    @patch_cli
    def test_get_all(self):
        disks = get_disks()
        assert_that(len(disks), equal_to(180))
        for disk in disks:
            if disk.index == '4_0_e8':
                verify_disk_4_0_e8(disk)
                break

    @patch_cli
    def test_get_disk(self):
        disk = VNXDisk.get(t_cli(), '4_0_e8')
        verify_disk_4_0_e8(disk)

    @property
    @instance_cache
    def disk_001(self):
        return VNXDisk('0_0_1', t_cli())

    @patch_cli
    def test_delete_disk(self):
        ret = self.disk_001.delete()
        assert_that(ret, has_items(''))

    @patch_cli
    def test_install_disk(self):
        ret = self.disk_001.install()
        assert_that(ret, has_items(''))

    @patch_cli
    def test_disk_read_iops(self):
        assert_that(self.disk_001.read_iops, equal_to(4.1))

    @patch_cli
    def test_disk_write_iops(self):
        assert_that(self.disk_001.write_iops, equal_to(5.1))

    @patch_cli
    def test_disk_total_iops(self):
        assert_that(self.disk_001.total_iops, equal_to(9.2))

    @patch_cli
    def test_disk_read_mbps(self):
        assert_that(self.disk_001.read_mbps, equal_to(3.2))

    @patch_cli
    def test_disk_write_mbps(self):
        assert_that(self.disk_001.write_mbps, equal_to(4.2))

    @patch_cli
    def test_disk_total_mbps(self):
        assert_that(self.disk_001.total_mbps, equal_to(7.4))

    @patch_cli
    def test_disk_utilization(self):
        assert_that(self.disk_001.utilization, equal_to(40.0))

    @patch_cli
    def test_disk_read_size_kb(self):
        assert_that(self.disk_001.read_size_kb, close_to(799.0, 1))

    @patch_cli
    def test_disk_write_size_kb(self):
        assert_that(self.disk_001.write_size_kb, close_to(844.0, 1))


class VNXDiskListTest(TestCase):
    disk_indices = ['0_0_C8', '4_0_D0', '4_0_E8']

    @patch_cli
    def test_all(self):
        disks = get_disks()
        assert_that(len(disks), equal_to(180))

    @patch_cli
    def test_index_filter(self):
        disks = get_disks().shadow_copy(disk_indices=self.disk_indices)
        assert_that(len(disks), equal_to(3))

    @patch_cli
    def test_multiple_filters(self):
        disks = get_disks().shadow_copy()
        disks.set_drive_type('NL SAS')
        assert_that(len(disks), equal_to(42))
        disks.set_capacity(2817564)
        assert_that(len(disks), equal_to(40))

    @patch_cli
    def test_get_same_disks_available(self):
        disks = get_disks().shadow_copy(disk_indices=self.disk_indices)
        disks.same_disks(2)
        assert_that(len(disks), equal_to(2))

    @patch_cli
    def test_get_same_disks_not_available(self):
        disks = get_disks().shadow_copy(disk_indices=self.disk_indices)
        disks.same_disks(3)
        assert_that(len(disks), equal_to(0))

    def test_resource_class_name(self):
        assert_that(VNXDiskList.resource_class_name(), equal_to('VNXDisk'))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import unicode_literals

import functools
import logging
import pipes

import six
from retryz import retry

from storops.connection import client
from storops.connection.exceptions import SFtpExecutionError, \
    SSHExecutionError, HTTPClientError
from storops.lib import common

paramiko = common.try_import('paramiko')

LOG = logging.getLogger(__name__)


def require_csrf_token(func):
    @functools.wraps(func)
    def decorator(self, url, **kwargs):
        wrapped = retry(on_error=self._http_authentication_error,
                        on_retry=self._update_csrf_token)(func)
        return wrapped(self, url, **kwargs)

    return decorator


class UnityRESTConnector(object):
    HEADERS = {
        'Accept': 'application/json',
        'Content-Type': 'application/json',
        'Accept_Language': 'en_US',
        'Visibility': 'Engineering',
        'X-EMC-REST-CLIENT': 'true',
        'User-agent': 'EMC-OpenStack',
    }

    def __init__(self, host, port=443, user='admin', password='',
                 verify=False, retries=None, cache_interval=0,
                 connect_timeout=30):
        base_url = 'https://{host}:{port}'.format(host=host, port=port)

        insecure = False
        ca_cert_path = None
        if isinstance(verify, bool):
            insecure = not verify
        else:
            ca_cert_path = verify
        self.http_client = client.HTTPClient(base_url=base_url,
                                             headers=self.HEADERS,
                                             auth=(user, password),
                                             insecure=insecure,
                                             retries=retries,
                                             ca_cert_path=ca_cert_path,
                                             cache_interval=cache_interval,
                                             timeout=(connect_timeout, None))

    def get(self, url, **kwargs):
        return self.http_client.get(url, **kwargs)

    @staticmethod
    def _http_authentication_error(err):
        return isinstance(err, HTTPClientError) and err.http_status == 401

    @require_csrf_token
    def post(self, url, **kwargs):
        return self.http_client.post(url, **kwargs)

    @require_csrf_token
    def delete(self, url, **kwargs):
        return self.http_client.delete(url, **kwargs)

    def _update_csrf_token(self):
        path_user = '/api/types/user/instances'
        resp, body = self.get(path_user)
        headers = {'emc-csrf-token': resp.headers['emc-csrf-token']}
        self.http_client.update_headers(headers)


class XMLAPIConnector(object):
    HEADERS = {
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    def __init__(self, host, user='nasadim', password=''):
        base_url = 'https://{}'.format(host)

        self.http_client = client.HTTPClient(base_url=base_url,
                                             headers=self.HEADERS,
                                             auth=(user, password),
                                             insecure=True)

        self.host = host
        self.user = user
        self.password = password

        self._login(host, user, password)

    def _login(self, host, user, password):
        url = 'https://{}/Login/?user={}&password={}&Login=Login'.format(
            host, user, password)

        self.http_client.request(url, 'GET')

    def post(self, body):
        url = '/servlets/CelerraManagementServices'
        try:
            return self.http_client.post(url, body=body)
        except HTTPClientError as ex:
            if ex.http_status == 403:
                self._login(self.host, self.user, self.password)
                self.http_client.post(url, body=body)
            else:
                raise


class SSHConnector(object):
    """SSH Connection to the specified host."""

    def __init__(self, host, username, password, port=22):
        self.transport = None
        self.init_connection(host, password, port, username)
        self.isLive = True

    def init_connection(self, host, password, port, username):
        self.transport = paramiko.Transport((host, port))
        # Currently we only support to use the password to ssh.
        if password:
            try:
                self.transport.connect(username=username, password=password)
            except paramiko.ssh_exception.SSHException as ex:
                error_msg = ('Failed to setup SSH connection. '
                             'Reason:%s.' % six.text_type(ex))
                LOG.error(error_msg)
                raise ex

    def execute(self, command, timeout=None, check_exit_code=True):
        cmd = ' '.join(pipes.quote(cmd_arg) for cmd_arg in command)
        channel = self.transport.open_session()
        channel.exec_command(cmd)
        channel.settimeout(timeout)
        exit_status = channel.recv_exit_status()

        stdout = channel.makefile('r').read()
        stderr = channel.makefile_stderr('r').read()

        channel.makefile('wb').close()
        self._ssh_command_log(cmd, stdout, stderr)

        # exit_status == -1 if no exit code was returned
        if exit_status != -1:
            LOG.debug('Result was %s' % exit_status)
            if check_exit_code and exit_status != 0:
                raise SSHExecutionError(exit_code=exit_status,
                                        stdout=stdout,
                                        stderr=stderr,
                                        cmd=cmd)

        return stdout, stderr

    def copy_file_to_remote(self, local_path, remote_path):
        """scp the local file to remote folder.

        :param local_path: local path
        :param remote_path: remote path
        """
        sftp_client = self.transport.open_sftp_client()
        LOG.debug('Copy the local file to remote. '
                  'Source=%(src)s. Target=%(target)s.' %
                  {'src': local_path, 'target': remote_path})
        try:
            sftp_client.put(local_path, remote_path)
        except Exception as ex:
            LOG.error('Failed to copy the local file to remote. '
                      'Reason: %s.' % six.text_type(ex))
            raise SFtpExecutionError(err=ex)

    def get_remote_file(self, remote_path, local_path):
        """Fetch remote File.

        :param remote_path: remote path
        :param local_path: local path
        """
        sftp_client = self.transport.open_sftp_client()
        LOG.debug('Get the remote file. '
                  'Source=%(src)s. Target=%(target)s.' %
                  {'src': remote_path, 'target': local_path})
        try:
            sftp_client.get(remote_path, local_path)
        except Exception as ex:
            LOG.error('Failed to secure copy. Reason: %s.' %
                      six.text_type(ex))
            raise SFtpExecutionError(err=ex)

    def close(self):
        """Closes the ssh connection."""
        if 'isLive' in self.__dict__ and self.isLive:
            self.transport.close()
            self.isLive = False

    def __del__(self):
        """Try to close the ssh connection if not explicitly closed."""
        self.close()

    @staticmethod
    def _ssh_command_log(command, stdout, stderr):
        LOG.debug('[SSH Command] {}\n'
                  '[stdout] {}\n'
                  '[stderr] {}\n'
                  .format(command, stdout, stderr))



# Copyright (c) 2018 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Veriion 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import unittest

import ddt
import hamcrest as hc

import storops
from storops.unity.resource import replication_interface, health, port, sp
from storops_test.unity import rest_mock as mock

__author__ = 'Ryan Liang'


@ddt.ddt
class UnityReplicationInterfaceTest(unittest.TestCase):
    @mock.patch_rest
    def test_get_properties(self):
        ri = replication_interface.UnityReplicationInterface.get(
            cli=mock.t_rest(), _id='if_10')
        hc.assert_that(ri.gateway, hc.equal_to('10.245.47.1'))
        hc.assert_that(ri.health, hc.instance_of(health.UnityHealth))
        hc.assert_that(ri.health.value, hc.equal_to(storops.HealthEnum.OK))
        hc.assert_that(ri.id, hc.equal_to('if_10'))
        hc.assert_that(ri.ip_address, hc.equal_to('10.245.47.99'))
        hc.assert_that(ri.ip_port, hc.instance_of(port.UnityIpPort))
        hc.assert_that(ri.ip_port.id, hc.equal_to('spa_eth2'))
        hc.assert_that(ri.ip_protocol_version,
                       hc.equal_to(storops.IpProtocolVersionEnum.IPv4))
        hc.assert_that(ri.mac_address, hc.equal_to('00:60:16:5C:07:0B'))
        hc.assert_that(ri.name, hc.equal_to('4_FNM00150600267'))
        hc.assert_that(ri.netmask, hc.equal_to('255.255.255.0'))
        hc.assert_that(ri.v6_prefix_length, hc.none())
        hc.assert_that(ri.vlan_id, hc.none())

    @mock.patch_rest
    def test_get_collection(self):
        ri_list = replication_interface.UnityReplicationInterfaceList.get(
            mock.t_rest())
        hc.assert_that(len(ri_list), hc.equal_to(2))
        hc.assert_that([ri.id for ri in ri_list],
                       hc.equal_to(['if_10', 'if_11']))

    @mock.patch_rest
    @ddt.data({'sp': sp.UnityStorageProcessor(_id='spa', cli=mock.t_rest()),
               'ip_port': port.UnityIpPort(_id='spa_eth3', cli=mock.t_rest()),
               'ip_address': '10.10.10.10', 'netmask': None,
               'v6_prefix_length': None, 'gateway': None, 'vlan_id': None},
              {'sp': sp.UnityStorageProcessor(_id='spa', cli=mock.t_rest()),
               'ip_port': port.UnityIpPort(_id='spa_eth3', cli=mock.t_rest()),
               'ip_address': '10.10.10.10', 'netmask': '255.255.255.0',
               'v6_prefix_length': None, 'gateway': '10.10.10.1',
               'vlan_id': 111})
    @ddt.unpack
    def test_create(self, sp, ip_port, ip_address, netmask, v6_prefix_length,
                    gateway, vlan_id):
        ri = replication_interface.UnityReplicationInterface.create(
            mock.t_rest(), sp, ip_port, ip_address, netmask=netmask,
            v6_prefix_length=v6_prefix_length, gateway=gateway,
            vlan_id=vlan_id)
        hc.assert_that(ri.id, hc.equal_to('if_10'))

    @mock.patch_rest
    @ddt.data({'sp': sp.UnityStorageProcessor(_id='spb', cli=mock.t_rest()),
               'ip_port': port.UnityIpPort(_id='spb_eth3', cli=mock.t_rest()),
               'ip_address': '10.10.20.10', 'netmask': '255.255.255.0',
               'v6_prefix_length': None, 'gateway': '10.10.20.1',
               'vlan_id': 222})
    @ddt.unpack
    def test_modify(self, sp, ip_port, ip_address, netmask, v6_prefix_length,
                    gateway, vlan_id):
        ri = replication_interface.UnityReplicationInterface.get(
            cli=mock.t_rest(), _id='if_10')
        resp = ri.modify(sp=sp, ip_port=ip_port, ip_address=ip_address,
                         netmask=netmask, v6_prefix_length=v6_prefix_length,
                         gateway=gateway, vlan_id=vlan_id)
        hc.assert_that(resp.is_ok(), hc.equal_to(True))



# coding=utf-8
# Copyright (c) 2015 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

import os
from unittest import TestCase

from hamcrest import assert_that, equal_to

from storops_test.utils import read_test_file
from storops_test.vnx.nas_mock import MockXmlPost

__author__ = 'Cedric Zhuang'


class NasMockTest(TestCase):
    def test_get_name(self):
        post_body = read_test_file(self.get_folder(), 'fs_single_post.xml')
        filename = MockXmlPost.get_filename(post_body)
        assert_that(filename,
                    equal_to('get_fs_abc.xml'))

    def get_folder(self):
        return os.path.join('vnx', 'testdata', 'nas_xml_output')

    def test_get_folder(self):
        post_body = read_test_file(self.get_folder(), 'fs_single_post.xml')
        folder = MockXmlPost.get_folder(post_body)
        expected = os.path.join(self.get_folder(), 'Query',
                                'FileSystemQueryParams')
        assert_that(folder, equal_to(expected))



# Copyright (c) 2016 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import unicode_literals

from unittest import TestCase

from hamcrest import assert_that, instance_of, equal_to, close_to

from storops.vnx.calculator import VNXMetricConfigParser, VNXMetricConfig, \
    VNXMetricConfigList, minus, div, round_60, add, aggregated_sum
from storops.vnx.resource.lun import VNXLun, VNXLunList
from storops_test.utils import is_nan
from storops_test.vnx.cli_mock import t_cli, patch_cli

__author__ = 'Cedric Zhuang'

NaN = float('nan')


class MathOperationsTest(TestCase):
    def test_add_left_nan(self):
        assert_that(add(NaN, 5.0), equal_to(5.0))

    def test_add_right_nan(self):
        assert_that(add(5.0, NaN), equal_to(5.0))

    def test_add_two_nan(self):
        assert_that(add(NaN, NaN), equal_to(0.0))

    def test_add_left_none(self):
        assert_that(add(None, 5.0), equal_to(5.0))

    def test_add_right_none(self):
        assert_that(add(5.0, None), equal_to(5.0))

    def test_add_two_none(self):
        assert_that(add(None, None), equal_to(0.0))

    def test_add_two_normal(self):
        assert_that(add(5.1, 6.2), equal_to(11.3))

    def test_minus_op1_none(self):
        assert_that(minus(None, 5.0), is_nan())

    def test_minus_op2_none(self):
        assert_that(minus(1, None), is_nan())

    def test_minus_normal(self):
        assert_that(minus(32, 50.0), equal_to(-18.0))

    def test_minus_nan(self):
        assert_that(minus(NaN, 5.0), is_nan())

    def test_div_op1_none(self):
        assert_that(div(None, 5.0), is_nan())

    def test_div_op2_none(self):
        assert_that(div(3, None), is_nan())

    def test_div_op2_zero(self):
        assert_that(div(4, 0), is_nan())

    def test_div_both_zero(self):
        assert_that(div(0, 0), equal_to(0.0))

    def test_div_nan(self):
        assert_that(div(float('nan'), 4.2), is_nan())

    def test_div_normal(self):
        assert_that(div(6, 2.0), equal_to(3.0))

    def test_round_60_higher_half(self):
        assert_that(round_60(113.5), equal_to(120))

    def test_round_60_lower_half(self):
        assert_that(round_60(132.3), equal_to(120))

    def test_round_60_nan(self):
        assert_that(round_60(NaN), is_nan())

    def test_round_60_none(self):
        assert_that(round_60(None), is_nan())


class AggregatorTest(TestCase):
    @patch_cli
    def test_aggregated_sum(self):
        lun_list = VNXLunList(cli=t_cli())
        ret = aggregated_sum(lun_list, 'consumed_capacity_gbs')
        assert_that(ret, close_to(4911.59, 0.01))


class VNXMetricConfigCalculatorTest(TestCase):
    def test_get_config(self):
        config = VNXMetricConfigParser.get_config(VNXLun)
        assert_that(config, instance_of(VNXMetricConfigList))

    def test_get_metric_config(self):
        config = VNXMetricConfigParser.get_config(VNXLun)
        metric_config = config.get_metric_config('read_iops')
        assert_that(metric_config, instance_of(VNXMetricConfig))
        assert_that(metric_config.name, equal_to('read_iops'))
